# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains acquisition function builders, which build and define our acquisition
functions --- functions that estimate the utility of evaluating sets of candidate points.
"""
from __future__ import annotations

from typing import Callable, Mapping, Optional, cast

import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import check_shapes

from ...data import Dataset
from ...models import ProbabilisticModel, ReparametrizationSampler
from ...models.interfaces import (
    HasReparamSampler,
    SupportsGetObservationNoise,
    SupportsReparamSamplerObservationNoise,
)
from ...space import SearchSpace
from ...types import Tag, TensorType
from ...utils import DEFAULTS
from ..interface import (
    AcquisitionFunction,
    AcquisitionFunctionBuilder,
    AcquisitionFunctionClass,
    ProbabilisticModelType,
    SingleModelAcquisitionBuilder,
    SingleModelVectorizedAcquisitionBuilder,
)
from .utils import MultivariateNormalCDF


class ProbabilityOfImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builder for the probability of improvement function, where the "best" value
    is taken to be the minimum of the posterior mean at observed points.
    """

    def __repr__(self) -> str:
        """"""
        return "ProbabilityOfImprovement()"

    def prepare_acquisition_function(
        self, model: ProbabilisticModel, dataset: Optional[Dataset] = None
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :return: The probability of improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)[0]
        return probability_below_threshold(model, eta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer.  Must be populated.
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(isinstance(function, probability_below_threshold), [tf.constant([])])
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)[0]
        function.update(eta)  # type: ignore
        return function


class ExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builder for the expected improvement function where the "best" value is taken to be the minimum
    of the posterior mean at observed points.

    In the presence of constraints in the search_space the "best" value is computed only at the
    feasible query points. If there are no feasible points, the "best" value is instead taken to be
    the maximum of the posterior mean at all observed points.
    """

    def __init__(self, search_space: Optional[SearchSpace] = None):
        """
        :param search_space: The global search space over which the optimisation is defined. This is
            only used to determine explicit constraints.
        """
        self._search_space = search_space

    def __repr__(self) -> str:
        """"""
        return f"ExpectedImprovement({self._search_space!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :return: The expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        # Check feasibility against any explicit constraints in the search space.
        if self._search_space is not None and self._search_space.has_constraints:
            is_feasible = self._search_space.is_feasible(dataset.query_points)
            if not tf.reduce_any(is_feasible):
                query_points = dataset.query_points
            else:
                query_points = tf.boolean_mask(dataset.query_points, is_feasible)
        else:
            is_feasible = tf.constant([True], dtype=bool)
            query_points = dataset.query_points

        mean, _ = model.predict(query_points)
        if not tf.reduce_any(is_feasible):
            eta = tf.reduce_max(mean, axis=0)
        else:
            eta = tf.reduce_min(mean, axis=0)

        return expected_improvement(model, eta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer.  Must be populated.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(isinstance(function, expected_improvement), [tf.constant([])])

        # Check feasibility against any explicit constraints in the search space.
        if self._search_space is not None and self._search_space.has_constraints:
            is_feasible = self._search_space.is_feasible(dataset.query_points)
            if not tf.reduce_any(is_feasible):
                query_points = dataset.query_points
            else:
                query_points = tf.boolean_mask(dataset.query_points, is_feasible)
        else:
            is_feasible = tf.constant([True], dtype=bool)
            query_points = dataset.query_points

        mean, _ = model.predict(query_points)
        if not tf.reduce_any(is_feasible):
            eta = tf.reduce_max(mean, axis=0)
        else:
            eta = tf.reduce_min(mean, axis=0)

        function.update(eta)  # type: ignore
        return function


class expected_improvement(AcquisitionFunctionClass):
    def __init__(self, model: ProbabilisticModel, eta: TensorType):
        r"""
        Return the Expected Improvement (EI) acquisition function for single-objective global
        optimization. Improvement is with respect to the current "best" observation ``eta``, where
        an improvement moves towards the objective function's minimum and the expectation is
        calculated with respect to the ``model`` posterior. For model posterior :math:`f`, this is

        .. math:: x \mapsto \mathbb E \left[ \max (\eta - f(x), 0) \right]

        This function was introduced by Mockus et al, 1975. See :cite:`Jones:1998` for details.

        :param model: The model of the objective function.
        :param eta: The "best" observation.
        :return: The expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        self._model = model
        self._eta = tf.Variable(eta)

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta value."""
        self._eta.assign(eta)

    @tf.function
    @check_shapes(
        "x: [N..., 1, D] # This acquisition function only supports batch sizes of one",
        "return: [N..., L]",
    )
    def __call__(self, x: TensorType) -> TensorType:
        mean, variance = self._model.predict(tf.squeeze(x, -2))
        normal = tfp.distributions.Normal(mean, tf.sqrt(variance))
        return (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)


class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[SupportsGetObservationNoise]):
    """
    Builder for the augmented expected improvement function for optimization single-objective
    optimization problems with high levels of observation noise.
    """

    def __repr__(self) -> str:
        """"""
        return "AugmentedExpectedImprovement()"

    def prepare_acquisition_function(
        self,
        model: SupportsGetObservationNoise,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :return: The expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        if not isinstance(model, SupportsGetObservationNoise):
            raise NotImplementedError(
                f"AugmentedExpectedImprovement only works with models that support "
                f"get_observation_noise; received {model!r}"
            )
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)
        return augmented_expected_improvement(model, eta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: SupportsGetObservationNoise,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(isinstance(function, augmented_expected_improvement), [tf.constant([])])
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)
        function.update(eta)  # type: ignore
        return function


class augmented_expected_improvement(AcquisitionFunctionClass):
    def __init__(self, model: SupportsGetObservationNoise, eta: TensorType):
        r"""
        Return the Augmented Expected Improvement (AEI) acquisition function for single-objective
        global optimization under homoscedastic observation noise.
        Improvement is with respect to the current "best" observation ``eta``, where an
        improvement moves towards the objective function's minimum and the expectation is calculated
        with respect to the ``model`` posterior. In contrast to standard EI, AEI has an additional
        multiplicative factor that penalizes evaluations made in areas of the space with very small
        posterior predictive variance. Thus, when applying standard EI to noisy optimisation
        problems, AEI avoids getting trapped and repeatedly querying the same point.
        For model posterior :math:`f`, this is
        .. math:: x \mapsto EI(x) * \left(1 - frac{\tau^2}{\sqrt{s^2(x)+\tau^2}}\right),
        where :math:`s^2(x)` is the predictive variance and :math:`\tau` is observation noise.
        This function was introduced by Huang et al, 2006. See :cite:`Huang:2006` for details.

        :param model: The model of the objective function.
        :param eta: The "best" observation.
        :return: The expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one or a model without homoscedastic observation noise.
        """
        self._model = model
        self._eta = tf.Variable(eta)
        self._noise_variance = tf.Variable(model.get_observation_noise())

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta value and noise variance."""
        self._eta.assign(eta)
        self._noise_variance.assign(self._model.get_observation_noise())

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        mean, variance = self._model.predict(tf.squeeze(x, -2))
        normal = tfp.distributions.Normal(mean, tf.sqrt(variance))
        ei = (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)
        augmentation = 1 - (tf.math.sqrt(self._noise_variance)) / (
            tf.math.sqrt(self._noise_variance + variance)
        )
        return ei * augmentation


class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builder for the negative of the lower confidence bound. The lower confidence bound is typically
    minimised, so the negative is suitable for maximisation.
    """

    def __init__(self, beta: float = 1.96):
        """
        :param beta: Weighting given to the variance contribution to the lower confidence bound.
            Must not be negative.
        """
        self._beta = beta

    def __repr__(self) -> str:
        """"""
        return f"NegativeLowerConfidenceBound({self._beta!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.
        :return: The negative lower confidence bound function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        :raise ValueError: If ``beta`` is negative.
        """
        lcb = lower_confidence_bound(model, self._beta)
        return tf.function(lambda at: -lcb(at))

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        return function  # no need to update anything


class NegativePredictiveMean(NegativeLowerConfidenceBound):
    """
    Builder for the negative of the predictive mean. The predictive mean is minimised on minimising
    the objective function. The negative predictive mean is therefore maximised.
    """

    def __init__(self) -> None:
        super().__init__(beta=0.0)

    def __repr__(self) -> str:
        """"""
        return "NegativePredictiveMean()"


def lower_confidence_bound(model: ProbabilisticModel, beta: float) -> AcquisitionFunction:
    r"""
    The lower confidence bound (LCB) acquisition function for single-objective global optimization.

    .. math:: x^* \mapsto \mathbb{E} [f(x^*)|x, y] - \beta \sqrt{ \mathrm{Var}[f(x^*)|x, y] }

    See :cite:`Srinivas:2010` for details.

    :param model: The model of the objective function.
    :param beta: The weight to give to the standard deviation contribution of the LCB. Must not be
        negative.
    :return: The lower confidence bound function. This function will raise
        :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
        greater than one.
    :raise tf.errors.InvalidArgumentError: If ``beta`` is negative.
    """
    tf.debugging.assert_non_negative(
        beta, message="Standard deviation scaling parameter beta must not be negative"
    )

    @tf.function
    def acquisition(x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        mean, variance = model.predict(tf.squeeze(x, -2))
        return mean - beta * tf.sqrt(variance)

    return acquisition


class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    r"""
    Uses the :func:`probability_below_threshold` function to build a
    probability of feasiblity acquisition function, defined in :cite:`gardner14` as

    .. math::

        \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d} c(\mathbf{x})
        \qquad ,

    where :math:`\tau` is a threshold. Values below the threshold are considered feasible by the
    constraint function. See also :cite:`schonlau1998global` for details.
    """

    def __init__(self, threshold: float | TensorType):
        """
        :param threshold: The (scalar) probability of feasibility threshold.
        :raise ValueError (or InvalidArgumentError): If ``threshold`` is not a scalar.
        """
        tf.debugging.assert_scalar(threshold)

        self._threshold = threshold

    def __repr__(self) -> str:
        """"""
        return f"ProbabilityOfFeasibility({self._threshold!r})"

    @property
    def threshold(self) -> float | TensorType:
        """The probability of feasibility threshold."""
        return self._threshold

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.
        :return: The probability of feasibility function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        return probability_below_threshold(model, self.threshold)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        return function  # no need to update anything


class probability_below_threshold(AcquisitionFunctionClass):
    def __init__(self, model: ProbabilisticModel, threshold: float | TensorType):
        r"""
        The probability of being below the threshold. This brings together commonality
        between probability of improvement and probability of feasiblity.
        Probability is is caculated with respect to the `model` posterior.
        For model posterior :math:`f`, this is
        .. math:: x \mapsto \mathbb P \left (f(x) < \eta)\right]
        where :math:`\eta` is the threshold.
        :param model: The model of the objective function.
        :param threshold: The (scalar) probability of feasibility threshold.
        :return: The probability of feasibility function. This function will raise
        :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
        greater than one.
        :raise ValueError or tf.errors.InvalidArgumentError: If ``threshold`` is not a scalar.
        """
        tf.debugging.assert_scalar(threshold)
        self._model = model
        self._threshold = tf.Variable(threshold)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        mean, var = self._model.predict(tf.squeeze(x, -2))
        distr = tfp.distributions.Normal(mean, tf.sqrt(var))
        return distr.cdf(tf.cast(self._threshold, x.dtype))

    def update(self, threshold: TensorType) -> None:
        """Update the acquisition function with a new threshold value."""
        self._threshold.assign(threshold)


class FastConstraintsFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builds a feasiblity acquisition function from the residuals of explicit constraints defined in
    the search space.
    """

    def __init__(
        self,
        search_space: SearchSpace,
        smoothing_function: Optional[Callable[[TensorType], TensorType]] = None,
    ):
        """
        :param search_space: The global search space over which the feasibility of the constraints
            is defined.
        :param smoothing_function: The smoothing function used for constraints residuals. The
            default is CDF of the Normal distribution with a scale of `1e-3`.
        :raise NotImplementedError: If the `search_space` does not have constraints.
        """
        if not search_space.has_constraints:
            raise NotImplementedError(
                "FastConstraintsFeasibility requires constraints in the search space."
            )

        self._search_space = search_space
        self._smoothing_function = smoothing_function

    def __repr__(self) -> str:
        """"""
        return f"FastConstraintsFeasibility({self._search_space!r}, {self._smoothing_function!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: Unused.
        :param dataset: Unused.
        :return: The function for feasibility of constraints.
        """
        return fast_constraints_feasibility(self._search_space, self._smoothing_function)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: Unused.
        :param dataset: Unused.
        :return: The function for feasibility of constraints.
        """
        return function  # No need to update anything.


def fast_constraints_feasibility(
    search_space: SearchSpace,
    smoothing_function: Optional[Callable[[TensorType], TensorType]] = None,
) -> AcquisitionFunction:
    """
    Returns a feasiblity acquisition function from the residuals of explicit constraints defined in
    the search space.

    :param search_space: The global search space over which the feasibility of the constraints
        is defined.
    :param smoothing_function: The smoothing function used for constraints residuals. The
        default is CDF of the Normal distribution with a scale of `1e-3`.
    :return: The function for feasibility of constraints.
    :raise NotImplementedError: If the `search_space` does not have constraints.
    """

    if not search_space.has_constraints:
        raise NotImplementedError(
            "fast_constraints_feasibility requires constraints in the search space."
        )

    @tf.function
    def acquisition(x: TensorType) -> TensorType:
        if smoothing_function is None:
            _smoothing_function = tfp.distributions.Normal(
                tf.constant(0.0, x.dtype), tf.constant(1e-3, x.dtype)
            ).cdf
        else:
            _smoothing_function = smoothing_function
        residuals = search_space.constraints_residuals(x)
        return tf.math.reduce_prod(_smoothing_function(residuals), axis=-1)

    return acquisition


class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticModelType]):
    """
    Builder for the *expected constrained improvement* acquisition function defined in
    :cite:`gardner14`. The acquisition function computes the expected improvement from the best
    feasible point, where feasible points are those that (probably) satisfy some constraint. Where
    there are no feasible points, this builder simply builds the constraint function.
    """

    def __init__(
        self,
        objective_tag: Tag,
        constraint_builder: AcquisitionFunctionBuilder[ProbabilisticModelType],
        min_feasibility_probability: float | TensorType = 0.5,
        search_space: Optional[SearchSpace] = None,
    ):
        """
        :param objective_tag: The tag for the objective data and model.
        :param constraint_builder: The builder for the constraint function.
        :param min_feasibility_probability: The minimum probability of feasibility for a
            "best point" to be considered feasible.
        :param search_space: The global search space over which the optimisation is defined. This is
            only used to determine explicit constraints.
        :raise ValueError (or tf.errors.InvalidArgumentError): If ``min_feasibility_probability``
            is not a scalar in the unit interval :math:`[0, 1]`.
        """
        tf.debugging.assert_scalar(min_feasibility_probability)

        if isinstance(min_feasibility_probability, (int, float)):
            tf.debugging.assert_greater_equal(float(min_feasibility_probability), 0.0)
            tf.debugging.assert_less_equal(float(min_feasibility_probability), 1.0)
        else:
            dtype = min_feasibility_probability.dtype
            tf.debugging.assert_greater_equal(min_feasibility_probability, tf.constant(0, dtype))
            tf.debugging.assert_less_equal(min_feasibility_probability, tf.constant(1, dtype))

        self._objective_tag = objective_tag
        self._constraint_builder = constraint_builder
        self._search_space = search_space
        self._min_feasibility_probability = min_feasibility_probability
        self._constraint_fn: Optional[AcquisitionFunction] = None
        self._expected_improvement_fn: Optional[AcquisitionFunction] = None
        self._constrained_improvement_fn: Optional[AcquisitionFunction] = None

    def __repr__(self) -> str:
        """"""
        return (
            f"ExpectedConstrainedImprovement({self._objective_tag!r}, {self._constraint_builder!r},"
            f" {self._min_feasibility_probability!r}, {self._search_space!r})"
        )

    def prepare_acquisition_function(
        self,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        """
        :param models: The models over each tag.
        :param datasets: The data from the observer.
        :return: The expected constrained improvement acquisition function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        :raise KeyError: If `objective_tag` is not found in ``datasets`` and ``models``.
        :raise tf.errors.InvalidArgumentError: If the objective data is empty.
        """
        tf.debugging.Assert(datasets is not None, [tf.constant([])])
        datasets = cast(Mapping[Tag, Dataset], datasets)

        objective_model = models[self._objective_tag]
        objective_dataset = datasets[self._objective_tag]

        tf.debugging.assert_positive(
            len(objective_dataset),
            message="Expected improvement is defined with respect to existing points in the"
            " objective data, but the objective data is empty.",
        )

        self._constraint_fn = self._constraint_builder.prepare_acquisition_function(
            models, datasets=datasets
        )
        pof = self._constraint_fn(objective_dataset.query_points[:, None, ...])
        is_feasible = tf.squeeze(pof >= self._min_feasibility_probability, axis=-1)

        # Check feasibility against any explicit constraints in the search space.
        if self._search_space is not None and self._search_space.has_constraints:
            ss_is_feasible = self._search_space.is_feasible(objective_dataset.query_points)
            is_feasible = tf.logical_and(is_feasible, ss_is_feasible)

        if not tf.reduce_any(is_feasible):
            return self._constraint_fn

        feasible_query_points = tf.boolean_mask(objective_dataset.query_points, is_feasible)
        feasible_mean, _ = objective_model.predict(feasible_query_points)
        self._update_expected_improvement_fn(objective_model, feasible_mean)

        @tf.function
        def constrained_function(x: TensorType) -> TensorType:
            return cast(AcquisitionFunction, self._expected_improvement_fn)(x) * cast(
                AcquisitionFunction, self._constraint_fn
            )(x)

        self._constrained_improvement_fn = constrained_function
        return constrained_function

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param models: The models for each tag.
        :param datasets: The data from the observer.
        """
        tf.debugging.Assert(datasets is not None, [tf.constant([])])
        datasets = cast(Mapping[Tag, Dataset], datasets)

        objective_model = models[self._objective_tag]
        objective_dataset = datasets[self._objective_tag]

        tf.debugging.assert_positive(
            len(objective_dataset),
            message="Expected improvement is defined with respect to existing points in the"
            " objective data, but the objective data is empty.",
        )
        tf.debugging.Assert(self._constraint_fn is not None, [tf.constant([])])

        constraint_fn = cast(AcquisitionFunction, self._constraint_fn)
        self._constraint_builder.update_acquisition_function(
            constraint_fn, models, datasets=datasets
        )
        pof = constraint_fn(objective_dataset.query_points[:, None, ...])
        is_feasible = tf.squeeze(pof >= self._min_feasibility_probability, axis=-1)

        # Check feasibility against any explicit constraints in the search space.
        if self._search_space is not None and self._search_space.has_constraints:
            ss_is_feasible = self._search_space.is_feasible(objective_dataset.query_points)
            is_feasible = tf.logical_and(is_feasible, ss_is_feasible)

        if not tf.reduce_any(is_feasible):
            return constraint_fn

        feasible_query_points = tf.boolean_mask(objective_dataset.query_points, is_feasible)
        feasible_mean, _ = objective_model.predict(feasible_query_points)
        self._update_expected_improvement_fn(objective_model, feasible_mean)

        if self._constrained_improvement_fn is not None:
            return self._constrained_improvement_fn

        @tf.function
        def constrained_function(x: TensorType) -> TensorType:
            return cast(AcquisitionFunction, self._expected_improvement_fn)(x) * cast(
                AcquisitionFunction, self._constraint_fn
            )(x)

        self._constrained_improvement_fn = constrained_function
        return self._constrained_improvement_fn

    def _update_expected_improvement_fn(
        self, objective_model: ProbabilisticModelType, feasible_mean: TensorType
    ) -> None:
        """
        Set or update the unconstrained expected improvement function.

        :param objective_model: The objective model.
        :param feasible_mean: The mean of the feasible query points.
        """
        eta = tf.reduce_min(feasible_mean, axis=0)

        if self._expected_improvement_fn is None:
            self._expected_improvement_fn = expected_improvement(objective_model, eta)
        else:
            tf.debugging.Assert(
                isinstance(self._expected_improvement_fn, expected_improvement), [tf.constant([])]
            )
            self._expected_improvement_fn.update(eta)  # type: ignore


class MonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[HasReparamSampler]):
    """
    Builder for a Monte Carlo-based expected improvement function for use with a model without
    analytical expected improvement (e.g. a deep GP). The "best" value is taken to be
    the minimum of the posterior mean at observed points. See
    :class:`monte_carlo_expected_improvement` for details.
    """

    def __init__(self, sample_size: int, *, jitter: float = DEFAULTS.JITTER):
        """
        :param sample_size: The number of samples for each batch of points.
        :param jitter: The jitter for the reparametrization sampler.
        :raise tf.errors.InvalidArgumentError: If ``sample_size`` is not positive, or ``jitter`` is
            negative.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_greater_equal(jitter, 0.0)

        super().__init__()

        self._sample_size = sample_size
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""
        return f"MonteCarloExpectedImprovement({self._sample_size!r}, jitter={self._jitter!r})"

    def prepare_acquisition_function(
        self,
        model: HasReparamSampler,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model over the specified ``dataset``. Must have output dimension [1].
        :param dataset: The data from the observer. Cannot be empty.
        :return: The estimated *expected improvement* acquisition function.
        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, ``model``
            does not have an output dimension of [1] or does not have a ``reparam_sample`` method.
        """
        if not isinstance(model, HasReparamSampler):
            raise ValueError(
                f"MonteCarloExpectedImprovement only supports models with a reparam_sampler method;"
                f"received {model!r}"
            )

        sampler = model.reparam_sampler(self._sample_size)

        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        samples_at_query_points = sampler.sample(
            dataset.query_points[..., None, :], jitter=self._jitter
        )
        mean = tf.reduce_mean(samples_at_query_points, axis=-3, keepdims=True)  # [N, 1, 1, L]

        tf.debugging.assert_shapes(
            [(mean, [..., 1])], message="Expected model with output dimension [1]."
        )

        eta = tf.squeeze(tf.reduce_min(mean, axis=0))

        return monte_carlo_expected_improvement(sampler, eta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: HasReparamSampler,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model. Must have output dimension [1]. Unused here.
        :param dataset: The data from the observer. Cannot be empty
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(
            isinstance(function, monte_carlo_expected_improvement), [tf.constant([])]
        )
        sampler = function._sampler  # type: ignore
        sampler.reset_sampler()
        samples_at_query_points = sampler.sample(
            dataset.query_points[..., None, :], jitter=self._jitter
        )
        mean = tf.reduce_mean(samples_at_query_points, axis=-3, keepdims=True)

        tf.debugging.assert_shapes(
            [(mean, [..., 1])], message="Expected model with output dimension [1]."
        )

        eta = tf.squeeze(tf.reduce_min(mean, axis=0))
        function.update(eta)  # type: ignore
        return function


class monte_carlo_expected_improvement(AcquisitionFunctionClass):
    r"""
    Return a Monte Carlo based Expected Improvement (EI) acquisition function for
    single-objective global optimization. Improvement is with respect to the current "best"
    observation ``eta``, where an improvement moves towards the objective function's minimum
    and the expectation is calculated with respect to the ``model`` posterior. For model
    posterior :math:`f`, this is

    .. math:: x \mapsto \mathbb E \left[ \max (\eta - f(x), 0) \right].

    For the Monte Carlo version, the expectation is calculated by samples that we save. See
    :cite:`wilson2018maximizing` for details.
    """

    def __init__(self, sampler: ReparametrizationSampler[HasReparamSampler], eta: TensorType):
        r"""
        :param sampler: The model sampler of the objective function.
        :param eta: The "best" observation.
        :return: The Monte Carlo expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        self._sampler = sampler
        self._eta = tf.Variable(eta)

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta value."""
        self._eta.assign(eta)

    @tf.function
    def __call__(self, at: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(at, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        samples = tf.squeeze(self._sampler.sample(at), axis=-1)  # [..., S, 1]
        improvement = tf.maximum(self._eta - samples, 0.0)  # [..., S, 1]
        return tf.reduce_mean(improvement, axis=-2)  # [..., 1]


class MonteCarloAugmentedExpectedImprovement(
    SingleModelAcquisitionBuilder[SupportsReparamSamplerObservationNoise]
):
    """
    Builder for a Monte Carlo-based augmented expected improvement function for use with a model
    without analytical augmented expected improvement (e.g. a deep GP). The "best" value is taken to
    be the minimum of the posterior mean at observed points. See
    :class:`monte_carlo_augmented_expected_improvement` for details.
    """

    def __init__(self, sample_size: int, *, jitter: float = DEFAULTS.JITTER):
        """
        :param sample_size: The number of samples for each batch of points.
        :param jitter: The jitter for the reparametrization sampler.
        :raise tf.errors.InvalidArgumentError: If ``sample_size`` is not positive, or ``jitter`` is
            negative.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_greater_equal(jitter, 0.0)

        super().__init__()

        self._sample_size = sample_size
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""
        return (
            f"MonteCarloAugmentedExpectedImprovement({self._sample_size!r}, "
            f"jitter={self._jitter!r})"
        )

    def prepare_acquisition_function(
        self,
        model: SupportsReparamSamplerObservationNoise,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model over the specified ``dataset``. Must have output dimension [1].
        :param dataset: The data from the observer. Cannot be empty.
        :return: The estimated *expected improvement* acquisition function.
        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, ``model``
            does not have an output dimension of [1], does not have a ``reparam_sample`` method, or
            does not support observation noise.
        """
        if not isinstance(model, SupportsReparamSamplerObservationNoise):
            raise ValueError(
                f"MonteCarloAugmentedExpectedImprovement only supports models with a "
                f"reparam_sampler method and that support observation noise; received "
                f"{model!r}."
            )

        sampler = model.reparam_sampler(self._sample_size)

        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        samples_at_query_points = sampler.sample(
            dataset.query_points[..., None, :], jitter=self._jitter
        )
        mean = tf.reduce_mean(samples_at_query_points, axis=-3, keepdims=True)  # [N, 1, 1, L]

        tf.debugging.assert_shapes(
            [(mean, [..., 1])], message="Expected model with output dimension [1]."
        )

        eta = tf.squeeze(tf.reduce_min(mean, axis=0))

        return monte_carlo_augmented_expected_improvement(model, sampler, eta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: SupportsReparamSamplerObservationNoise,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model. Must have output dimension [1]. Unused here
        :param dataset: The data from the observer. Cannot be empty.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(
            isinstance(function, monte_carlo_augmented_expected_improvement), [tf.constant([])]
        )
        sampler = function._sampler  # type: ignore
        sampler.reset_sampler()
        samples_at_query_points = sampler.sample(
            dataset.query_points[..., None, :], jitter=self._jitter
        )
        mean = tf.reduce_mean(samples_at_query_points, axis=-3, keepdims=True)  # [N, 1, 1, L]

        tf.debugging.assert_shapes(
            [(mean, [..., 1])], message="Expected model with output dimension [1]."
        )

        eta = tf.squeeze(tf.reduce_min(mean, axis=0))
        function.update(eta)  # type: ignore
        return function


class monte_carlo_augmented_expected_improvement(AcquisitionFunctionClass):
    r"""
    Return a Monte Carlo based Augmented Expected Improvement (AEI) acquisition function for
    single-objective global optimization with high levels of observation noise. See
    :cite:`wilson2018maximizing` for details on using the reparametrization trick for optimizing
    acquisition functions and :cite:`Huang:2006`: for details of AEI.
    """

    def __init__(
        self,
        model: SupportsReparamSamplerObservationNoise,
        sampler: ReparametrizationSampler[SupportsReparamSamplerObservationNoise],
        eta: TensorType,
    ):
        r"""
        :param model: The model of the objective function.
        :param sampler: The model sampler of the objective function.
        :param eta: The "best" observation.
        :return: The Monte Carlo expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        self._model = model
        self._sampler = sampler
        self._eta = tf.Variable(eta)
        self._noise_variance = tf.Variable(model.get_observation_noise())

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta and noise variance"""
        self._eta.assign(eta)
        self._noise_variance.assign(self._model.get_observation_noise())

    @tf.function
    def __call__(self, at: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(at, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        samples = tf.squeeze(self._sampler.sample(at), axis=-1)  # [..., S, 1]
        improvement = tf.maximum(self._eta - samples, 0.0)  # [..., S, 1]
        variance = tf.math.reduce_variance(samples, -2)  # [..., 1]
        augmentation = 1 - (
            tf.math.sqrt(self._noise_variance) / tf.math.sqrt(self._noise_variance + variance)
        )
        return augmentation * tf.reduce_mean(improvement, axis=-2)  # [..., 1]


class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[HasReparamSampler]):
    """
    Expected improvement for batches of points (or :math:`q`-EI), approximated using Monte Carlo
    estimation with the reparametrization trick. See :cite:`Ginsbourger2010` for details.
    Improvement is measured with respect to the minimum predictive mean at observed query points.
    This is calculated in :class:`BatchMonteCarloExpectedImprovement` by assuming observations
    at new points are independent from those at known query points. This is faster, but is an
    approximation for noisy observers.
    """

    def __init__(self, sample_size: int, *, jitter: float = DEFAULTS.JITTER):
        """
        :param sample_size: The number of samples for each batch of points.
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
        :raise tf.errors.InvalidArgumentError: If ``sample_size`` is not positive, or ``jitter``
            is negative.
        """
        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_greater_equal(jitter, 0.0)

        self._sample_size = sample_size
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""
        return f"BatchMonteCarloExpectedImprovement({self._sample_size!r}, jitter={self._jitter!r})"

    def prepare_acquisition_function(
        self,
        model: HasReparamSampler,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model. Must have event shape [1].
        :param dataset: The data from the observer. Must be populated.
        :return: The batch *expected improvement* acquisition function.
        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, or ``model``
            does not have an event shape of [1].
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        mean, _ = model.predict(dataset.query_points)

        tf.debugging.assert_shapes(
            [(mean, ["_", 1])], message="Expected model with event shape [1]."
        )

        eta = tf.reduce_min(mean, axis=0)
        return batch_monte_carlo_expected_improvement(self._sample_size, model, eta, self._jitter)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: HasReparamSampler,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model. Must have event shape [1].
        :param dataset: The data from the observer. Must be populated.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(
            isinstance(function, batch_monte_carlo_expected_improvement), [tf.constant([])]
        )
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)
        function.update(eta)  # type: ignore
        return function


class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
    def __init__(self, sample_size: int, model: HasReparamSampler, eta: TensorType, jitter: float):
        """
        :param sample_size: The number of Monte-Carlo samples.
        :param model: The model of the objective function.
        :param eta: The "best" observation.
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
        :return: The expected improvement function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        self._sample_size = sample_size

        if not isinstance(model, HasReparamSampler):
            raise ValueError(
                f"The batch Monte-Carlo expected improvement acquisition function only supports "
                f"models that implement a reparam_sampler method; received {model!r}"
            )

        sampler = model.reparam_sampler(self._sample_size)

        self._sampler = sampler
        self._eta = tf.Variable(eta)
        self._jitter = jitter

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta value and reset the reparam sampler."""
        self._eta.assign(eta)
        self._sampler.reset_sampler()

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        samples = tf.squeeze(self._sampler.sample(x, jitter=self._jitter), axis=-1)  # [..., S, B]
        min_sample_per_batch = tf.reduce_min(samples, axis=-1)  # [..., S]
        batch_improvement = tf.maximum(self._eta - min_sample_per_batch, 0.0)  # [..., S]
        return tf.reduce_mean(batch_improvement, axis=-1, keepdims=True)  # [..., 1]


class BatchExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """Accurate approximation of the batch expected improvement, using the
    method of Chvallier and Ginsbourger :cite:`chevalier2013fast`.

    Internally, this uses a highly accurate approximation of the cumulative
    density function of the multivariate Gaussian, developed by Alan Genz
    :cite:`genz2016numerical`.
    """

    def __init__(
        self,
        sample_size: int,
        *,
        jitter: float = DEFAULTS.JITTER,
    ):
        """Initialise the BatchExpectedImprovement instance.

        :param sample_size: int, number of Sobol samples to use.
        :param jitter: float, amount of jitter for Cholesky factorisations.
        """

        tf.debugging.assert_positive(sample_size)
        tf.debugging.assert_greater_equal(jitter, 0.0)

        self._sample_size = sample_size
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""

        return f"BatchExpectedImprovement({self._sample_size!r}, jitter={self._jitter!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model. Must have event shape [1].
        :param dataset: The data from the observer. Must be populated.
        :return: The batch *expected improvement* acquisition function.
        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, or ``model``
            does not have an event shape of [1].
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        # Get mean and covariance
        mean, _ = model.predict(dataset.query_points)

        tf.debugging.assert_shapes(
            [(mean, ["_", 1])],
            message="Expected model with event shape [1].",
        )

        eta = tf.reduce_min(mean, axis=0)

        acquisition_function = batch_expected_improvement(
            self._sample_size,
            model,
            eta,
            self._jitter,
        )

        return acquisition_function

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model. Must have event shape [1].
        :param dataset: The data from the observer. Must be populated.
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(isinstance(function, batch_expected_improvement), [])

        # Get mean and covariance
        mean, _ = model.predict(dataset.query_points)
        eta = tf.reduce_min(mean, axis=0)

        function.update(eta=eta)  # type: ignore

        return function


class batch_expected_improvement(AcquisitionFunctionClass):
    def __init__(
        self,
        sample_size: int,
        model: ProbabilisticModel,
        eta: TensorType,
        jitter: float,
    ):
        """Initialise the batch_expected_improvement instance.

        :param sample_size: int, number of samples to use.
        :param model: Gaussian process regression model.
        :param eta: Tensor of shape (,), expected improvement threshold. This
            is the best value observed so far durin the BO loop.
        :param jitter: float, amount of jitter for Cholesky factorisations.
        """

        self._sample_size = sample_size
        self._jitter = jitter
        self._eta = tf.Variable(eta)
        self._model = model

        self._mvn_cdf_1: Optional[MultivariateNormalCDF] = None
        self._mvn_cdf_2: Optional[MultivariateNormalCDF] = None

        self._num_sobol_skip = int(tf.math.floor(10**9 * tf.random.uniform((), dtype=tf.float32)))

    def update(self, eta: TensorType) -> None:
        """Update the acquisition function with a new eta value and reset the
        reparam sampler.
        """
        self._eta.assign(eta)
        self._num_sobol_skip = int(tf.math.floor(10**9 * tf.random.uniform((), dtype=tf.float32)))

    def _compute_bm(
        self,
        mean: tf.Tensor,
        threshold: tf.Tensor,
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensors b and m as detailed in Chevalier and Ginsbourger
        :cite:`chevalier2013fast`.

        :param mean: Tensor of shape (B, Q)
        :param threshold: Tensor of shape (B,)

        :returns b: Tensor of shape (B, Q, Q)
        :returns m: Tensor of shape (B, Q, Q)
        """

        # Check shapes of input tensors
        tf.debugging.assert_shapes(
            [
                (mean, ("B", "Q")),
                (threshold, ("B",)),
            ]
        )

        # Unpack tensor shape and data type
        B, Q = mean.shape
        dtype = mean.dtype

        # Compute b tensor
        threshold = tf.tile(threshold[:, None], (1, Q))
        threshold = tf.linalg.diag(threshold)  # (B, Q, Q)

        b = tf.zeros(shape=(B, Q, Q), dtype=dtype)
        b = b - threshold

        # Compute m tensor
        m = mean[:, None, :] - mean[:, :, None]  # (B, Q, Q)
        m = m - tf.linalg.diag(mean)  # (B, Q, Q)

        return b, m

    def _delta(
        self, idx: int, dim: int, B: int, transpose: bool, dtype: tf.DType
    ) -> TensorType:  # pragma: no cover (tf.map_fn)
        """Helper function for the _compute_Sigma function, which computes a
        *delta* tensor of shape (B, idx, idx) such that

            delta[B, i, :] = 1 if i == idx
            delta[B, i, :] = 0 otherwise.

        If transpose == True, then the last two dimensions of the tensor are
        transposed, in which case

            delta[B, :, i] = 1 if i == idx
            delta[B, :, i] = 0 otherwise.

        :param idx: Index for entries equal to 1.
        :param dim: Dimension of the last and second to last axes.
        :param B: Leading dimension of tensor.
        :param transpose: Whether to transpose the last two dimensions or not.
        :param dtype: The dtype of the tensor, either tf.float32 or tf.float64.
        """

        # Check input parameters
        tf.debugging.assert_non_negative(idx)
        tf.debugging.assert_non_negative(dim)
        tf.debugging.assert_positive(B)

        o1 = tf.ones(shape=(B, idx, dim), dtype=dtype)
        z1 = tf.zeros(shape=(B, 1, dim), dtype=dtype)
        o2 = tf.ones(shape=(B, dim - idx - 1, dim), dtype=dtype)

        delta = tf.concat([o1, z1, o2], axis=1)
        delta = tf.transpose(delta, perm=[0, 2, 1]) if transpose else delta

        return delta

    def _compute_Sigma(
        self,
        covariance: tf.Tensor,
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensor Sigma, as detailed in Chevalier and Ginsbourger
        :cite:`chevalier2013fast`.

        :param covariance: Tensor of shape (B, Q, Q)
        :returns Sigma: Tensor of shape (B, Q, Q, Q)
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes([(covariance, ("B", "Q", "Q"))])

        # Unpack tensor shape and dtype
        B, Q, _ = covariance.shape
        dtype = covariance.dtype

        Sigma = tf.zeros(shape=(B, Q, Q, Q))

        def compute_single_slice(q: int) -> TensorType:  # pragma: no cover (tf.map_fn)
            diq = self._delta(q, Q, B, transpose=False, dtype=dtype)
            dqj = self._delta(q, Q, B, transpose=True, dtype=dtype)

            Sigma_ij = covariance[:, :, :]
            Sigma_iq = covariance[:, :, q : q + 1]
            Sigma_qj = covariance[:, q : q + 1, :]
            Sigma_qq = covariance[:, q : q + 1, q : q + 1]

            cov = Sigma_ij * diq * dqj - Sigma_iq * diq - Sigma_qj * dqj + Sigma_qq

            return cov

        Sigma = tf.map_fn(
            compute_single_slice,
            tf.range(Q),
            fn_output_signature=dtype,
        )

        Sigma = tf.transpose(Sigma, perm=[1, 0, 2, 3])

        return Sigma

    def _compute_p(
        self,
        m_reshaped: tf.Tensor,
        b_reshaped: tf.Tensor,
        Sigma_reshaped: tf.Tensor,
        mvn_cdf: Callable[[TensorType, TensorType, TensorType, float], TensorType],
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensor p, as detailed in Chevalier and Ginsbourger
        :cite:`chevalier2013fast`.

        :param m_reshaped: Tensor of shape (BQ, Q)
        :param b_reshaped: Tensor of shape (BQ, Q)
        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
        :returns p: Tensor of shape (B, Q)
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes(
            [
                (m_reshaped, ("BQ", "Q")),
                (b_reshaped, ("BQ", "Q")),
                (Sigma_reshaped, ("BQ", "Q", "Q")),
            ]
        )

        # Unpack dtype and mean shape
        dtype = m_reshaped.dtype
        BQ, Q = m_reshaped.shape  # (B*Q, Q)

        if BQ % Q == 0:
            B = BQ // Q

        else:
            raise ValueError(
                f"Expected size of dimension 0 of m_reshaped tensor to be "
                f"divisible by size of dimension 1, instead found "
                f"{m_reshaped.shape[0]} and {m_reshaped.shape[1]}."
            )

        # Compute mean, covariance and x for p mvn normal cdf
        p_cdf_mean = tf.zeros(shape=(BQ, Q), dtype=dtype)  # (B*Q, Q)
        p_cdf_cov = Sigma_reshaped  # (B*Q, Q, Q)

        p_cdf_x = b_reshaped - m_reshaped  # (B*Q, Q)

        p = mvn_cdf(  # type: ignore
            x=p_cdf_x,
            mean=p_cdf_mean,
            cov=p_cdf_cov,
        )  # (B*Q,)

        p = tf.reshape(p, shape=(B, Q))  # (B, Q)

        return p

    def _compute_c(
        self,
        m_reshaped: tf.Tensor,
        b_reshaped: tf.Tensor,
        Sigma_reshaped: tf.Tensor,
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensor c, which is the c^{(i)} tensor detailed in Chevalier and
        Ginsbourger :cite:`chevalier2013fast`.

        :param m_reshaped: Tensor of shape (BQ, Q)
        :param b_reshaped: Tensor of shape (BQ, Q)
        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
        :returns c: Tensor of shape (B, Q, Q-1)
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes(
            [
                (m_reshaped, ("BQ", "Q")),
                (b_reshaped, ("BQ", "Q")),
                (Sigma_reshaped, ("BQ", "Q", "Q")),
            ]
        )

        # Unpack tensor shape
        BQ, Q = m_reshaped.shape

        # Compute difference between b and m tensors
        diff = b_reshaped - m_reshaped  # (B*Q, Q)

        # Compute c, including the ith entry, which we want to remove
        cov_ratio = Sigma_reshaped / tf.linalg.diag_part(Sigma_reshaped)[:, :, None]  # (B*Q, Q, Q)
        c = diff[:, None, :] - diff[:, :, None] * cov_ratio  # (B*Q, Q, Q)

        # Remove the ith entry by masking c with a boolean mask with False across
        # the diagonal and True in the off-diagonal terms
        mask = tf.math.logical_not(tf.cast(tf.eye(Q, dtype=tf.int32), dtype=tf.bool))
        mask = tf.tile(mask[None, :, :], (c.shape[0], 1, 1))

        c = tf.ragged.boolean_mask(c, mask).to_tensor()

        return c

    def _compute_R(
        self,
        Sigma_reshaped: tf.Tensor,
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensor R, which is the Sigma^{(i)} tensor detailed in Chevalier
        and Ginsbourger :cite:`chevalier2013fast`.

        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
        :returns R: Tensor of shape (B, Q-1, Q-1)
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes([(Sigma_reshaped, ("BQ", "Q", "Q"))])

        # Unpack tensor shape
        BQ, Q, _ = Sigma_reshaped.shape

        Sigma_uv = tf.tile(Sigma_reshaped[:, None, :, :], (1, Q, 1, 1))
        Sigma_iu = tf.tile(Sigma_reshaped[:, :, :, None], (1, 1, 1, Q))
        Sigma_iv = tf.tile(Sigma_reshaped[:, :, None, :], (1, 1, Q, 1))
        Sigma_ii = tf.linalg.diag_part(Sigma_reshaped)[:, :, None, None]

        R_whole = Sigma_uv - Sigma_iu * Sigma_iv / Sigma_ii

        def create_blocks(q: int) -> TensorType:  # pragma: no cover (tf.map_fn)
            block1 = tf.concat(
                [
                    R_whole[:, q, :q, :q],
                    R_whole[:, q, q + 1 :, :q],
                ],
                axis=1,
            )

            block2 = tf.concat(
                [
                    R_whole[:, q, :q, q + 1 :],
                    R_whole[:, q, q + 1 :, q + 1 :],
                ],
                axis=1,
            )

            R_block = tf.concat([block1, block2], axis=2)

            return R_block

        R = tf.map_fn(
            create_blocks,
            tf.range(Q),
            fn_output_signature=R_whole.dtype,
        )
        R = tf.transpose(R, perm=[1, 0, 2, 3])

        return R

    def _compute_Phi(
        self,
        c: tf.Tensor,
        R: tf.Tensor,
        mvn_cdf: Callable[[TensorType, TensorType, TensorType, float], TensorType],
    ) -> TensorType:
        """Helper function for the batch expected improvement, which computes
        the tensor Phi, which is the tensor of multivariate Gaussian CDFs, in
        the inner sum of the equation (3) in Chevalier and Ginsbourger
        :cite:`chevalier2013fast`.

        :param c: Tensor of shape (BQ, Q, Q-1).
        :param R: Tensor of shape (BQ, Q, Q-1, Q-1).
        :param mvn_cdf: Multivariate Gaussian CDF, made using MultivariateNormalCDF.
        :returns Phi: Tensor of multivariate Gaussian CDFs.
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes(
            [
                (c, ("BQ", "Q", "Q_")),
                (R, ("BQ", "Q", "Q_", "Q_")),
            ]
        )

        # Unpack tensor shape and data type
        BQ, Q, _, Q_ = R.shape
        dtype = R.dtype

        try:
            assert BQ % Q == 0

        except AssertionError:
            raise ValueError(
                f"Expected size of dimension 0 of R tensor to be "
                f"divisible by size of dimension 1, instead found "
                f"{R.shape[0]} and {R.shape[1]}."
            )

        # Compute parallelisation dimension from batch size
        B = BQ // Q

        c_reshaped = tf.reshape(c, (BQ * Q, Q - 1))
        R_reshaped = tf.reshape(R, (BQ * Q, Q - 1, Q - 1))

        # Compute mean, covariance and x for Phi mvn normal cdf
        Phi_cdf_x = c_reshaped  # (B*Q, Q-1)
        Phi_cdf_mean = tf.zeros(shape=(BQ * Q, Q - 1), dtype=dtype)  # (B*Q*Q, Q)
        Phi_cdf_cov = R_reshaped  # (B*Q*Q, Q-1, Q-1)

        # Compute multivariate cdfs
        mvn_cdfs = mvn_cdf(  # type: ignore
            x=Phi_cdf_x,
            mean=Phi_cdf_mean,
            cov=Phi_cdf_cov,
        )
        mvn_cdfs = tf.reshape(mvn_cdfs, (B, Q, Q))  # (B, Q, Q)

        return mvn_cdfs

    def _compute_batch_expected_improvement(
        self,
        mean: tf.Tensor,
        covariance: tf.Tensor,
        threshold: tf.Tensor,
        mvn_cdf_1: Callable[[TensorType, TensorType, TensorType, float], TensorType],
        mvn_cdf_2: Callable[[TensorType, TensorType, TensorType, float], TensorType],
    ) -> TensorType:
        """Accurate Monte Carlo approximation of the batch expected
        improvement, using the method of Chevalier and Ginsbourger
        :cite:`chevalier2013fast`.

        :param mean: Tensor of shape (B, Q).
        :param covariance: Tensor of shape (B, Q, Q).
        :param threshold: Tensor of shape (B, Q).
        :param mvn_cdf_1: Callable computing the multivariate CDF of a Q-dimensional Gaussian.
        :param mvn_cdf_2: Callable computing the multivariate CDF of a (Q-1)-dimensional Gaussian.
        :returns ei: Tensor of shape (B,), expected improvement.
        """

        # Check shapes of covariance tensor
        tf.debugging.assert_shapes(
            [
                (mean, ("B", "Q")),
                (covariance, ("B", "Q", "Q")),
                (threshold, ("B",)),
            ]
        )

        # Unpack and mean shape
        B, Q = mean.shape

        # Compute b and m tensors
        b, m = self._compute_bm(
            mean=mean,
            threshold=threshold,
        )  # (B, Q, Q), (B, Q, Q)

        # Compute Sigma
        Sigma = self._compute_Sigma(covariance=covariance)  # (B, Q, Q, Q)

        # Reshape all tensors, for batching
        b_reshaped = tf.reshape(b, (B * Q, Q))
        m_reshaped = tf.reshape(m, (B * Q, Q))
        Sigma_reshaped = tf.reshape(Sigma, (B * Q, Q, Q))

        # Compute p tensor
        p = self._compute_p(
            m_reshaped=m_reshaped,
            b_reshaped=b_reshaped,
            Sigma_reshaped=Sigma_reshaped,
            mvn_cdf=mvn_cdf_1,
        )

        # Compute c
        c = self._compute_c(
            m_reshaped=m_reshaped,
            b_reshaped=b_reshaped,
            Sigma_reshaped=Sigma_reshaped,
        )  # (B*Q, Q, Q-1)

        # Compute Sigma_i
        R = self._compute_R(
            Sigma_reshaped=Sigma_reshaped,
        )  # (B*Q, Q, Q-1, Q-1)

        # Compute Q-1 multivariate CDFs
        Phi_mvn_cdfs = self._compute_Phi(
            c=c,
            R=R,
            mvn_cdf=mvn_cdf_2,
        )

        # Compute univariate pdfs
        S_diag = tf.linalg.diag_part(Sigma)
        normal = tfp.distributions.Normal(loc=m, scale=S_diag**0.5)
        uvn_pdfs = tf.math.exp(normal.log_prob(b))  # (B, Q, Q)

        Sigma_diag = tf.linalg.diag_part(tf.transpose(Sigma, perm=[0, 2, 1, 3]))
        Sigma_diag = tf.transpose(Sigma_diag, perm=[0, 2, 1])

        T = tf.tile(threshold[:, None], (1, Q))

        mean_T_term = (mean - T) * p

        # Compute inner sum
        sum_term = tf.reduce_sum(
            Sigma_diag * uvn_pdfs * Phi_mvn_cdfs,
            axis=2,
        )

        # Compute outer sum
        ei = tf.reduce_sum(mean_T_term + sum_term, axis=1)

        return ei

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        """Computes the accurate approximation of the multi-point expected
        improvement.

        :param x: Tensor of shape (B, Q, D).
        :returns ei: Tensor of shape (B,), expected improvement.
        """

        if self._mvn_cdf_1 is None:
            self._mvn_cdf_1 = MultivariateNormalCDF(
                sample_size=self._sample_size,
                dim=x.shape[1],
                dtype=x.dtype,
                num_sobol_skip=self._num_sobol_skip,
            )

        if self._mvn_cdf_2 is None:
            self._mvn_cdf_2 = MultivariateNormalCDF(
                sample_size=self._sample_size,
                dim=x.shape[1] - 1,
                dtype=x.dtype,
                num_sobol_skip=self._num_sobol_skip,
            )

        mean, covariance = self._model.predict_joint(x)  # type: ignore

        mean = mean[:, :, 0]
        covariance = covariance[:, 0, :, :]
        covariance = (
            covariance
            + 1e-6
            * tf.eye(
                covariance.shape[-1],
                dtype=covariance.dtype,
            )[None, :, :]
        )

        threshold = tf.tile(self._eta, (mean.shape[0],))

        # Check shapes of x, mean, covariance and threshold tensors
        tf.debugging.assert_shapes(
            [
                (x, ("B", "Q", "D")),
                (mean, ("B", "Q")),
                (covariance, ("B", "Q", "Q")),
                (threshold, ("B",)),
            ]
        )

        ei = self._compute_batch_expected_improvement(
            mean=-mean,
            covariance=covariance,
            threshold=-threshold,
            mvn_cdf_1=self._mvn_cdf_1,
            mvn_cdf_2=self._mvn_cdf_2,
        )[:, None]

        return ei


class MultipleOptimismNegativeLowerConfidenceBound(
    SingleModelVectorizedAcquisitionBuilder[ProbabilisticModel]
):
    """
    A simple parallelization of the lower confidence bound acquisition function that produces
    a vectorized acquisition function which can efficiently optimized even for large batches.

    See :cite:`torossian2020bayesian` for details.
    """

    def __init__(self, search_space: SearchSpace):
        """
        :param search_space: The global search space over which the optimisation is defined.
        """
        self._search_space = search_space

    def __repr__(self) -> str:
        """"""
        return f"MultipleOptimismNegativeLowerConfidenceBound({self._search_space!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.
        :return: The multiple optimism negative lower confidence bound function.
        """
        return multiple_optimism_lower_confidence_bound(model, self._search_space.dimension)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        tf.debugging.Assert(
            isinstance(function, multiple_optimism_lower_confidence_bound), [tf.constant([])]
        )
        return function  # nothing to update


class multiple_optimism_lower_confidence_bound(AcquisitionFunctionClass):
    r"""
    The multiple optimism lower confidence bound (MOLCB) acquisition function for single-objective
    global optimization.

    Each batch dimension of this acquisiton function correponds to a lower confidence bound
    acquisition function with different beta values, i.e. each point in a batch chosen by this
    acquisition function lies on a gradient of exploration/exploitation trade-offs.

    We choose the different beta values following the cdf method of :cite:`torossian2020bayesian`.
    See their paper for more details.
    """

    def __init__(self, model: ProbabilisticModel, search_space_dim: int):
        """
        :param model: The model of the objective function.
        :param search_space_dim: The dimensions of the optimisation problem's search space.
        :raise tf.errors.InvalidArgumentError: If ``search_space_dim`` is not postive.
        """

        tf.debugging.assert_positive(search_space_dim)
        self._search_space_dim = search_space_dim

        self._model = model
        self._initialized = tf.Variable(False)  # Keep track of when we need to resample
        self._betas = tf.Variable(tf.ones([0], dtype=tf.float64), shape=[None])  # [0] lazy init

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        batch_size = tf.shape(x)[-2]
        tf.debugging.assert_positive(batch_size)

        if self._initialized:  # check batch size hasnt changed during BO
            tf.debugging.assert_equal(
                batch_size,
                tf.shape(self._betas)[0],
                f"{type(self).__name__} requires a fixed batch size. Got batch size {batch_size}"
                f" but previous batch size was {tf.shape(self._betas)[0]}.",
            )

        if not self._initialized:
            normal = tfp.distributions.Normal(
                tf.constant(0.0, dtype=x.dtype), tf.constant(1.0, dtype=x.dtype)
            )
            spread = 0.5 + 0.5 * tf.range(1, batch_size + 1, dtype=x.dtype) / (
                tf.cast(batch_size, dtype=x.dtype) + 1.0
            )  # [B]
            betas = normal.quantile(spread)  # [B]
            scaled_betas = 5.0 * tf.cast(self._search_space_dim, dtype=x.dtype) * betas  # [B]
            self._betas.assign(scaled_betas)  # [B]
            self._initialized.assign(True)

        mean, variance = self._model.predict(x)  # [..., B, 1]
        mean, variance = tf.squeeze(mean, -1), tf.squeeze(variance, -1)
        return -mean + tf.sqrt(variance) * self._betas  # [..., B]


class MakePositive(SingleModelAcquisitionBuilder[ProbabilisticModelType]):
    r"""
    Converts an acquisition function builder into one that only returns positive values, via
    :math:`x \mapsto \log(1 + \exp(x))`.

    This is sometimes a useful transformation: for example, converting non-batch acquisition
    functions into batch acquisition functions with local penalization requires functions
    that only return positive values.
    """

    def __init__(
        self,
        base_acquisition_function_builder: SingleModelAcquisitionBuilder[ProbabilisticModelType],
    ) -> None:
        """
        :param base_acquisition_function_builder: Base acquisition function to be made positive.
        """
        self._base_builder = base_acquisition_function_builder

    def __repr__(self) -> str:
        """"""
        return f"MakePositive({self._base_builder})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data to use to build the acquisition function (optional).
        :return: An acquisition function.
        """
        self._base_function = self._base_builder.prepare_acquisition_function(model, dataset)

        @tf.function
        def acquisition(x: TensorType) -> TensorType:
            return tf.math.log(1 + tf.math.exp(self._base_function(x)))

        return acquisition

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer (optional).
        :return: The updated acquisition function.
        """
        up_fn = self._base_builder.update_acquisition_function(self._base_function, model, dataset)
        if up_fn is self._base_function:
            return function
        else:
            self._base_function = up_fn

            @tf.function
            def acquisition(x: TensorType) -> TensorType:
                return tf.math.log(1 + tf.math.exp(self._base_function(x)))

            return acquisition



# Copyright 2023 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains synthetic multi-fidelity objective functions, useful for experimentation.
"""
from dataclasses import dataclass

import numpy as np
import tensorflow as tf

from ..space import Box, DiscreteSearchSpace, SearchSpace, SearchSpaceType, TaggedProductSearchSpace
from ..types import TensorType
from .single_objectives import SingleObjectiveTestProblem


@dataclass(frozen=True)
class SingleObjectiveMultifidelityTestProblem(SingleObjectiveTestProblem[SearchSpaceType]):
    num_fidelities: int
    """The number of fidelities of test function"""

    fidelity_search_space: TaggedProductSearchSpace
    """The search space including fidelities"""


def linear_multifidelity(x: TensorType) -> TensorType:
    x_input = x[..., :-1]
    x_fidelity = x[..., -1:]

    f = 0.5 * ((6.0 * x_input - 2.0) ** 2) * tf.math.sin(12.0 * x_input - 4.0) + 10.0 * (
        x_input - 1.0
    )
    f = f + x_fidelity * (f - 20.0 * (x_input - 1.0))

    return f


_LINEAR_MULTIFIDELITY_MINIMIZERS = {
    2: tf.constant([[0.75724875]], tf.float64),
    3: tf.constant([[0.76333767]], tf.float64),
    5: tf.constant([[0.76801846]], tf.float64),
}


_LINEAR_MULTIFIDELITY_MINIMA = {
    2: tf.constant([-6.020740055], tf.float64),
    3: tf.constant([-6.634287061], tf.float64),
    5: tf.constant([-7.933019704], tf.float64),
}


def _linear_multifidelity_search_space_builder(
    n_fidelities: int, input_search_space: SearchSpace
) -> TaggedProductSearchSpace:
    fidelity_search_space = DiscreteSearchSpace(np.arange(n_fidelities, dtype=float).reshape(-1, 1))
    search_space = TaggedProductSearchSpace(
        [input_search_space, fidelity_search_space], ["input", "fidelity"]
    )
    return search_space


Linear2Fidelity = SingleObjectiveMultifidelityTestProblem(
    name="Linear 2 Fidelity",
    objective=linear_multifidelity,
    search_space=Box(np.zeros(1), np.ones(1)),
    fidelity_search_space=_linear_multifidelity_search_space_builder(
        2, Box(np.zeros(1), np.ones(1))
    ),
    minimizers=_LINEAR_MULTIFIDELITY_MINIMIZERS[2],
    minimum=_LINEAR_MULTIFIDELITY_MINIMA[2],
    num_fidelities=2,
)

Linear3Fidelity = SingleObjectiveMultifidelityTestProblem(
    name="Linear 3 Fidelity",
    objective=linear_multifidelity,
    search_space=Box(np.zeros(1), np.ones(1)),
    fidelity_search_space=_linear_multifidelity_search_space_builder(
        3, Box(np.zeros(1), np.ones(1))
    ),
    minimizers=_LINEAR_MULTIFIDELITY_MINIMIZERS[3],
    minimum=_LINEAR_MULTIFIDELITY_MINIMA[3],
    num_fidelities=3,
)

Linear5Fidelity = SingleObjectiveMultifidelityTestProblem(
    name="Linear 5 Fidelity",
    objective=linear_multifidelity,
    search_space=Box(np.zeros(1), np.ones(1)),
    fidelity_search_space=_linear_multifidelity_search_space_builder(
        5, Box(np.zeros(1), np.ones(1))
    ),
    minimizers=_LINEAR_MULTIFIDELITY_MINIMIZERS[5],
    minimum=_LINEAR_MULTIFIDELITY_MINIMA[5],
    num_fidelities=5,
)



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains local penalization-based acquisition function builders.
"""
from __future__ import annotations

from typing import Callable, Dict, Mapping, Optional, Union, cast

import gpflow
import tensorflow as tf
import tensorflow_probability as tfp
from check_shapes import check_shapes
from typing_extensions import Protocol, runtime_checkable

from ...data import Dataset
from ...models import FastUpdateModel, ModelStack, ProbabilisticModel
from ...models.interfaces import (
    PredictJointModelStack,
    PredictJointPredictYModelStack,
    PredictYModelStack,
    SupportsGetKernel,
    SupportsGetObservationNoise,
    SupportsPredictJoint,
    SupportsPredictY,
)
from ...observer import OBJECTIVE
from ...space import SearchSpace
from ...types import Tag, TensorType
from ..interface import (
    AcquisitionFunction,
    AcquisitionFunctionBuilder,
    GreedyAcquisitionFunctionBuilder,
    PenalizationFunction,
    SingleModelAcquisitionBuilder,
    SingleModelGreedyAcquisitionBuilder,
    UpdatablePenalizationFunction,
)
from .entropy import MinValueEntropySearch
from .function import ExpectedImprovement, MakePositive, expected_improvement


class LocalPenalization(SingleModelGreedyAcquisitionBuilder[ProbabilisticModel]):
    r"""
    Builder of the acquisition function maker for greedily collecting batches by local
    penalization.  The resulting :const:`AcquisitionFunctionMaker` takes in a set of pending
    points and returns a base acquisition function penalized around those points.
    An estimate of the objective function's Lipschitz constant is used to control the size
    of penalization.

    Local penalization allows us to perform batch Bayesian optimization with a standard (non-batch)
    acquisition function. All that we require is that the acquisition function takes strictly
    positive values. By iteratively building a batch of points though sequentially maximizing
    this acquisition function but down-weighted around locations close to the already
    chosen (pending) points, local penalization provides diverse batches of candidate points.

    Local penalization is applied to the acquisition function multiplicatively. However, to
    improve numerical stability, we perform additive penalization in a log space.

    The Lipschitz constant and additional penalization parameters are estimated once
    when first preparing the acquisition function with no pending points. These estimates
    are reused for all subsequent function calls.
    """

    def __init__(
        self,
        search_space: SearchSpace,
        num_samples: int = 500,
        penalizer: Optional[
            Callable[
                [ProbabilisticModel, TensorType, TensorType, TensorType],
                Union[PenalizationFunction, UpdatablePenalizationFunction],
            ]
        ] = None,
        base_acquisition_function_builder: (
            ExpectedImprovement
            | MinValueEntropySearch[ProbabilisticModel]
            | MakePositive[ProbabilisticModel]
            | None
        ) = None,
    ):
        """
        :param search_space: The global search space over which the optimisation is defined.
        :param num_samples: Size of the random sample over which the Lipschitz constant
            is estimated. We recommend scaling this with search space dimension.
        :param penalizer: The chosen penalization method (defaults to soft penalization). This
            should be a function that accepts a model, pending points, lipschitz constant and eta
            and returns a PenalizationFunction.
        :param base_acquisition_function_builder: Base acquisition function to be
            penalized (defaults to expected improvement). Local penalization only supports
            strictly positive acquisition functions.
        :raise tf.errors.InvalidArgumentError: If ``num_samples`` is not positive.
        """
        tf.debugging.assert_positive(num_samples)

        self._search_space = search_space
        self._num_samples = num_samples

        self._lipschitz_penalizer = soft_local_penalizer if penalizer is None else penalizer

        if base_acquisition_function_builder is None:
            self._base_builder: SingleModelAcquisitionBuilder[ProbabilisticModel] = (
                ExpectedImprovement()
            )
        else:
            self._base_builder = base_acquisition_function_builder

        self._lipschitz_constant = None
        self._eta = None
        self._base_acquisition_function: Optional[AcquisitionFunction] = None
        self._penalization: Optional[PenalizationFunction | UpdatablePenalizationFunction] = None
        self._penalized_acquisition: Optional[AcquisitionFunction] = None

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :param pending_points: The points we penalize with respect to.
        :return: The (log) expected improvement penalized with respect to the pending points.
        :raise tf.errors.InvalidArgumentError: If the ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        acq = self._update_base_acquisition_function(dataset, model)
        if pending_points is not None and len(pending_points) != 0:
            acq = self._update_penalization(acq, dataset, model, pending_points)

        return acq

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, of to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: The updated acquisition function.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(self._base_acquisition_function is not None, [tf.constant([])])

        if new_optimization_step:
            self._update_base_acquisition_function(dataset, model)

        if pending_points is None or len(pending_points) == 0:
            # no penalization required if no pending_points
            return cast(AcquisitionFunction, self._base_acquisition_function)

        return self._update_penalization(function, dataset, model, pending_points)

    def _update_penalization(
        self,
        function: Optional[AcquisitionFunction],
        dataset: Dataset,
        model: ProbabilisticModel,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        tf.debugging.assert_rank(pending_points, 2)

        if self._penalized_acquisition is not None and isinstance(
            self._penalization, UpdatablePenalizationFunction
        ):
            # if possible, just update the penalization function variables
            self._penalization.update(pending_points, self._lipschitz_constant, self._eta)
            return self._penalized_acquisition
        else:
            # otherwise construct a new penalized acquisition function
            self._penalization = self._lipschitz_penalizer(
                model, pending_points, self._lipschitz_constant, self._eta
            )
            self._penalized_acquisition = PenalizedAcquisition(
                cast(PenalizedAcquisition, self._base_acquisition_function), self._penalization
            )
            return self._penalized_acquisition

    @tf.function(experimental_relax_shapes=True)
    def _get_lipschitz_estimate(
        self, model: ProbabilisticModel, sampled_points: TensorType
    ) -> tuple[TensorType, TensorType]:
        with tf.GradientTape() as g:
            g.watch(sampled_points)
            mean, _ = model.predict(sampled_points)
        grads = g.gradient(mean, sampled_points)
        grads_norm = tf.norm(grads, axis=1)
        max_grads_norm = tf.reduce_max(grads_norm)
        eta = tf.reduce_min(mean, axis=0)
        return max_grads_norm, eta

    def _update_base_acquisition_function(
        self, dataset: Dataset, model: ProbabilisticModel
    ) -> AcquisitionFunction:
        samples = self._search_space.sample(num_samples=self._num_samples)
        samples = tf.concat([dataset.query_points, samples], 0)

        lipschitz_constant, eta = self._get_lipschitz_estimate(model, samples)
        if lipschitz_constant < 1e-5:  # threshold to improve numerical stability for 'flat' models
            lipschitz_constant = 10

        self._lipschitz_constant = lipschitz_constant
        self._eta = eta

        if self._base_acquisition_function is not None:
            self._base_acquisition_function = self._base_builder.update_acquisition_function(
                self._base_acquisition_function,
                model,
                dataset=dataset,
            )
        elif isinstance(self._base_builder, ExpectedImprovement):  # reuse eta estimate
            self._base_acquisition_function = cast(
                AcquisitionFunction, expected_improvement(model, self._eta)
            )
        else:
            self._base_acquisition_function = self._base_builder.prepare_acquisition_function(
                model,
                dataset=dataset,
            )
        return self._base_acquisition_function


class PenalizedAcquisition:
    """Class representing a penalized acquisition function."""

    # (note that this needs to be defined as a top level class make it pickleable)
    def __init__(
        self, base_acquisition_function: AcquisitionFunction, penalization: PenalizationFunction
    ):
        """
        :param base_acquisition_function: Base (unpenalized) acquisition function.
        :param penalization: Penalization function.
        """
        self._base_acquisition_function = base_acquisition_function
        self._penalization = penalization

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        log_acq = tf.math.log(self._base_acquisition_function(x)) + tf.math.log(
            self._penalization(x)
        )
        return tf.math.exp(log_acq)


class local_penalizer(UpdatablePenalizationFunction):
    def __init__(
        self,
        model: ProbabilisticModel,
        pending_points: TensorType,
        lipschitz_constant: TensorType,
        eta: TensorType,
    ):
        """Initialize the local penalizer.

        :param model: The model over the specified ``dataset``.
        :param pending_points: The points we penalize with respect to.
        :param lipschitz_constant: The estimated Lipschitz constant of the objective function.
        :param eta: The estimated global minima.
        :return: The local penalization function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one."""
        self._model = model

        mean_pending, variance_pending = model.predict(pending_points)
        self._pending_points = tf.Variable(pending_points, shape=[None, *pending_points.shape[1:]])
        self._radius = tf.Variable(
            tf.transpose((mean_pending - eta) / lipschitz_constant),
            shape=[1, None],
        )
        self._scale = tf.Variable(
            tf.transpose(tf.sqrt(variance_pending) / lipschitz_constant),
            shape=[1, None],
        )

    def update(
        self,
        pending_points: TensorType,
        lipschitz_constant: TensorType,
        eta: TensorType,
    ) -> None:
        """Update the local penalizer with new variable values."""
        mean_pending, variance_pending = self._model.predict(pending_points)
        self._pending_points.assign(pending_points)
        self._radius.assign(tf.transpose((mean_pending - eta) / lipschitz_constant))
        self._scale.assign(tf.transpose(tf.sqrt(variance_pending) / lipschitz_constant))


class soft_local_penalizer(local_penalizer):
    r"""
    Return the soft local penalization function used for single-objective greedy batch Bayesian
    optimization in :cite:`Gonzalez:2016`.

    Soft penalization returns the probability that a candidate point does not belong
    in the exclusion zones of the pending points. For model posterior mean :math:`\mu`, model
    posterior variance :math:`\sigma^2`, current "best" function value :math:`\eta`, and an
    estimated Lipschitz constant :math:`L`,the penalization from a set of pending point
    :math:`x'` on a candidate point :math:`x` is given by
    .. math:: \phi(x, x') = \frac{1}{2}\textrm{erfc}(-z)
    where :math:`z = \frac{1}{\sqrt{2\sigma^2(x')}}(L||x'-x|| + \eta - \mu(x'))`.

    The penalization from a set of pending points is just product of the individual
    penalizations. See :cite:`Gonzalez:2016` for a full derivation.

    :param model: The model over the specified ``dataset``.
    :param pending_points: The points we penalize with respect to.
    :param lipschitz_constant: The estimated Lipschitz constant of the objective function.
    :param eta: The estimated global minima.
    :return: The local penalization function. This function will raise
        :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
        greater than one.
    """

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This penalization function cannot be calculated for batches of points.",
        )

        pairwise_distances = tf.norm(
            tf.expand_dims(x, 1) - tf.expand_dims(self._pending_points, 0), axis=-1
        )
        standardised_distances = (pairwise_distances - self._radius) / self._scale

        normal = tfp.distributions.Normal(tf.constant(0, x.dtype), tf.constant(1, x.dtype))
        penalization = normal.cdf(standardised_distances)
        return tf.reduce_prod(penalization, axis=-1)


class hard_local_penalizer(local_penalizer):
    r"""
    Return the hard local penalization function used for single-objective greedy batch Bayesian
    optimization in :cite:`Alvi:2019`.

    Hard penalization is a stronger penalizer than soft penalization and is sometimes more effective
    See :cite:`Alvi:2019` for details. Our implementation follows theirs, with the penalization from
    a set of pending points being the product of the individual penalizations.

    :param model: The model over the specified ``dataset``.
    :param pending_points: The points we penalize with respect to.
    :param lipschitz_constant: The estimated Lipschitz constant of the objective function.
    :param eta: The estimated global minima.
    :return: The local penalization function. This function will raise
        :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
        greater than one.
    """

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This penalization function cannot be calculated for batches of points.",
        )

        pairwise_distances = tf.norm(
            tf.expand_dims(x, 1) - tf.expand_dims(self._pending_points, 0), axis=-1
        )

        p = -5  # following experiments of :cite:`Alvi:2019`.
        penalization = ((pairwise_distances / (self._radius + self._scale)) ** p + 1) ** (1 / p)
        return tf.reduce_prod(penalization, axis=-1)


@runtime_checkable
class FantasizerModelType(
    FastUpdateModel,
    SupportsPredictJoint,
    SupportsPredictY,
    SupportsGetKernel,
    SupportsGetObservationNoise,
    Protocol,
):
    """The model requirements for the Fantasizer acquisition function."""


class FantasizerModelStack(
    PredictJointModelStack, PredictYModelStack, ModelStack[FantasizerModelType]
):
    """
    A stack of models :class:`FantasizerModelType` models. Note that this delegates predict_joint
    and predict_y but none of the other methods.
    """


FantasizerModelOrStack = Union[FantasizerModelType, FantasizerModelStack]


class Fantasizer(GreedyAcquisitionFunctionBuilder[FantasizerModelOrStack]):
    r"""
    Builder of the acquisition function maker for greedily collecting batches.
    Fantasizer allows us to perform batch Bayesian optimization with any
    standard (non-batch) acquisition function.

    Here, every time a query point is chosen by maximising an acquisition function,
    its corresponding observation is "fantasized", and the models are conditioned further
    on this new artificial data.

    This implies that the models need to predict what their updated predictions would be given
    new data, see :class:`~FastUpdateModel`. These equations are for instance in closed form
    for the GPR model, see :cite:`chevalier2014corrected` (eqs. 8-10) for details.

    There are several ways to "fantasize" data: the "kriging believer" heuristic (KB, see
    :cite:`ginsbourger2010kriging`) uses the mean of the model as observations.
    "sample" uses samples from the model.
    """

    def __init__(
        self,
        base_acquisition_function_builder: Optional[
            AcquisitionFunctionBuilder[SupportsPredictJoint]
            | SingleModelAcquisitionBuilder[SupportsPredictJoint]
        ] = None,
        fantasize_method: str = "KB",
    ):
        """

        :param base_acquisition_function_builder: The acquisition function builder to use.
            Defaults to :class:`~trieste.acquisition.ExpectedImprovement`.
        :param fantasize_method: The following options are available: "KB" and "sample".
            See class docs for more details.
        :raise tf.errors.InvalidArgumentError: If ``fantasize_method`` is not "KB" or "sample".
        """
        tf.debugging.Assert(fantasize_method in ["KB", "sample"], [tf.constant([])])

        if base_acquisition_function_builder is None:
            base_acquisition_function_builder = ExpectedImprovement()

        if isinstance(base_acquisition_function_builder, SingleModelAcquisitionBuilder):
            base_acquisition_function_builder = base_acquisition_function_builder.using(OBJECTIVE)

        self._builder = base_acquisition_function_builder
        self._fantasize_method = fantasize_method

        self._base_acquisition_function: Optional[AcquisitionFunction] = None
        self._fantasized_acquisition: Optional[AcquisitionFunction] = None
        self._fantasized_models: Mapping[
            Tag, _fantasized_model | ModelStack[SupportsPredictJoint]
        ] = {}

    def _update_base_acquisition_function(
        self,
        models: Mapping[Tag, FantasizerModelOrStack],
        datasets: Optional[Mapping[Tag, Dataset]],
    ) -> AcquisitionFunction:
        if self._base_acquisition_function is not None:
            self._base_acquisition_function = self._builder.update_acquisition_function(
                self._base_acquisition_function, models, datasets
            )
        else:
            self._base_acquisition_function = self._builder.prepare_acquisition_function(
                models, datasets
            )
        return self._base_acquisition_function

    def _update_fantasized_acquisition_function(
        self,
        models: Mapping[Tag, FantasizerModelOrStack],
        datasets: Optional[Mapping[Tag, Dataset]],
        pending_points: TensorType,
    ) -> AcquisitionFunction:
        tf.debugging.assert_rank(pending_points, 2)

        fantasized_data = {
            tag: _generate_fantasized_data(
                fantasize_method=self._fantasize_method,
                model=model,
                pending_points=pending_points,
            )
            for tag, model in models.items()
        }

        if datasets is None:
            datasets = fantasized_data
        else:
            datasets = {tag: data + fantasized_data[tag] for tag, data in datasets.items()}

        if self._fantasized_acquisition is None:
            self._fantasized_models = {
                tag: _generate_fantasized_model(model, fantasized_data[tag])
                for tag, model in models.items()
            }
            self._fantasized_acquisition = self._builder.prepare_acquisition_function(
                cast(Dict[Tag, SupportsPredictJoint], self._fantasized_models), datasets
            )
        else:
            for tag, model in self._fantasized_models.items():
                if isinstance(model, ModelStack):
                    observations = tf.split(
                        fantasized_data[tag].observations, model._event_sizes, axis=-1
                    )
                    for submodel, obs in zip(model._models, observations):
                        submodel.update_fantasized_data(
                            Dataset(fantasized_data[tag].query_points, obs)
                        )

                else:
                    model.update_fantasized_data(fantasized_data[tag])
            self._builder.update_acquisition_function(
                self._fantasized_acquisition,
                cast(Dict[Tag, SupportsPredictJoint], self._fantasized_models),
                datasets,
            )

        return self._fantasized_acquisition

    def prepare_acquisition_function(
        self,
        models: Mapping[Tag, FantasizerModelOrStack],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        """

        :param models: The models over each tag.
        :param datasets: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :return: An acquisition function.
        """
        for model in models.values():
            if not (
                isinstance(model, FantasizerModelType)
                or isinstance(model, ModelStack)
                and all(isinstance(m, FantasizerModelType) for m in model._models)
            ):
                raise NotImplementedError(
                    f"Fantasizer only works with FastUpdateModel models that also support "
                    f"predict_joint, get_kernel and get_observation_noise, or with "
                    f"ModelStack stacks of such models; received {model!r}"
                )
        if pending_points is None:
            return self._update_base_acquisition_function(models, datasets)
        else:
            return self._update_fantasized_acquisition_function(models, datasets, pending_points)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        models: Mapping[Tag, FantasizerModelOrStack],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param models: The models over each tag.
        :param datasets: The data from the observer (optional).
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, of to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: The updated acquisition function.
        """
        if pending_points is None:
            return self._update_base_acquisition_function(models, datasets)
        else:
            return self._update_fantasized_acquisition_function(models, datasets, pending_points)


def _generate_fantasized_data(
    fantasize_method: str, model: FantasizerModelOrStack, pending_points: TensorType
) -> Dataset:
    """
    Generates "fantasized" data at pending_points depending on the chosen heuristic:
    - KB (kriging believer) uses the mean prediction of the models
    - sample uses samples from the GP posterior.

    :param fantasize_method: the following options are available: "KB" and "sample".
    :param model: a model with predict method
    :param dataset: past data
    :param pending_points: points at which to fantasize data
    :return: a fantasized dataset
    """
    if fantasize_method == "KB":
        fantasized_obs, _ = model.predict(pending_points)
    elif fantasize_method == "sample":
        fantasized_obs = model.sample(pending_points, num_samples=1)[0]
    else:
        raise NotImplementedError(f"fantasize_method must be KB or sample, received {model!r}")

    return Dataset(pending_points, fantasized_obs)


def _generate_fantasized_model(
    model: FantasizerModelOrStack, fantasized_data: Dataset
) -> _fantasized_model | PredictJointPredictYModelStack:
    if isinstance(model, ModelStack):
        observations = tf.split(fantasized_data.observations, model._event_sizes, axis=-1)
        fmods = []
        for mod, obs, event_size in zip(model._models, observations, model._event_sizes):
            fmods.append(
                (
                    _fantasized_model(mod, Dataset(fantasized_data.query_points, obs)),
                    event_size,
                )
            )
        return PredictJointPredictYModelStack(*fmods)
    else:
        return _fantasized_model(model, fantasized_data)


class _fantasized_model(
    SupportsPredictJoint, SupportsGetKernel, SupportsGetObservationNoise, SupportsPredictY
):
    """
    Creates a new model from an existing one and additional data.
    This new model posterior is conditioned on both current model data and the additional one.
    """

    def __init__(self, model: FantasizerModelType, fantasized_data: Dataset):
        """
        :param model: a model, must be of class `FastUpdateModel`
        :param fantasized_data: additional dataset to condition on
        :raise NotImplementedError: If model is not of class `FastUpdateModel`.
        """

        self._model = model
        self._fantasized_query_points = tf.Variable(
            fantasized_data.query_points,
            trainable=False,
            shape=[None, *fantasized_data.query_points.shape[1:]],
        )
        self._fantasized_observations = tf.Variable(
            fantasized_data.observations,
            trainable=False,
            shape=[None, *fantasized_data.observations.shape[1:]],
        )

    def update_fantasized_data(self, fantasized_data: Dataset) -> None:
        """
        :param fantasized_data: new additional dataset to condition on
        """
        self._fantasized_query_points.assign(fantasized_data.query_points)
        self._fantasized_observations.assign(fantasized_data.observations)

    @check_shapes(
        "query_points: [batch..., N, D]",
        "return[0]: [batch..., ..., N, L]",
        "return[1]: [batch..., ..., N, L]",
    )
    def predict(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        This function wraps conditional_predict_f. It cannot directly call
        conditional_predict_f, since it does not accept query_points with rank > 2.
        We use map_fn to allow leading dimensions for query_points.

        :param query_points: shape [...*, N, d]
        :return: mean, shape [...*, ..., N, L] and cov, shape [...*, ..., N, L],
            where ... are the leading dimensions of fantasized_data
        """

        def fun(qp: TensorType) -> tuple[TensorType, TensorType]:  # pragma: no cover (tf.map_fn)
            fantasized_data = Dataset(
                self._fantasized_query_points.value(), self._fantasized_observations.value()
            )
            return self._model.conditional_predict_f(qp, fantasized_data)

        return _broadcast_predict(query_points, fun)

    @check_shapes(
        "query_points: [batch..., N, D]",
        "return[0]: [batch..., ..., N, L]",
        "return[1]: [batch..., ..., L, N, N]",
    )
    def predict_joint(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        This function wraps conditional_predict_joint. It cannot directly call
        conditional_predict_joint, since it does not accept query_points with rank > 2.
        We use map_fn to allow leading dimensions for query_points.

        :param query_points: shape [...*, N, D]
        :return: mean, shape [...*, ..., N, L] and cov, shape [...*, ..., L, N, N],
            where ... are the leading dimensions of fantasized_data
        """

        def fun(qp: TensorType) -> tuple[TensorType, TensorType]:  # pragma: no cover (tf.map_fn)
            fantasized_data = Dataset(
                self._fantasized_query_points.value(), self._fantasized_observations.value()
            )
            return self._model.conditional_predict_joint(qp, fantasized_data)

        return _broadcast_predict(query_points, fun)

    @check_shapes(
        "query_points: [batch..., N, D]",
        "return: [batch..., ..., S, N, L]",
    )
    def sample(self, query_points: TensorType, num_samples: int) -> TensorType:
        """
        This function wraps conditional_predict_f_sample. It cannot directly call
        conditional_predict_joint, since it does not accept query_points with rank > 2.
        We use map_fn to allow leading dimensions for query_points.

        :param query_points: shape [...*, N, D]
        :param num_samples: number of samples.
        :return: samples of shape [...*, ..., S, N, L], where ... are the leading
            dimensions of fantasized_data
        """
        leading_dim, query_points_flatten = _get_leading_dim_and_flatten(query_points)
        # query_points_flatten: [B, n, d], leading_dim =...*, product = B

        samples = tf.map_fn(
            fn=lambda qp: self._model.conditional_predict_f_sample(
                qp,
                Dataset(
                    self._fantasized_query_points.value(), self._fantasized_observations.value()
                ),
                num_samples,
            ),
            elems=query_points_flatten,
        )  # [B, ..., S, L]
        return _restore_leading_dim(samples, leading_dim)

    @check_shapes(
        "query_points: [broadcast batch..., N, D]",
        "return[0]: [batch..., ..., N, L]",
        "return[1]: [batch..., ..., N, L]",
    )
    def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """
        This function wraps conditional_predict_y. It cannot directly call
        conditional_predict_joint, since it does not accept query_points with rank > 2.
        We use tf.map_fn to allow leading dimensions for query_points.

        :param query_points: shape [...*, N, D]
        :return: mean, shape [...*, ..., N, L] and var, shape [...*, ..., N, L],
            where ... are the leading dimensions of fantasized_data
        """

        def fun(qp: TensorType) -> tuple[TensorType, TensorType]:  # pragma: no cover (tf.map_fn)
            fantasized_data = Dataset(
                self._fantasized_query_points.value(), self._fantasized_observations.value()
            )
            return self._model.conditional_predict_y(qp, fantasized_data)

        return _broadcast_predict(query_points, fun)

    def get_observation_noise(self) -> TensorType:
        return self._model.get_observation_noise()

    def get_kernel(self) -> gpflow.kernels.Kernel:
        return self._model.get_kernel()

    def log(self, dataset: Optional[Dataset] = None) -> None:
        return self._model.log(dataset)


def _broadcast_predict(
    query_points: TensorType, fun: Callable[[TensorType], tuple[TensorType, TensorType]]
) -> tuple[TensorType, TensorType]:
    """
    Utility function that allows leading dimensions for query_points when
    fun only accepts rank 2 tensors. It works by flattening query_points into
    a rank 3 tensor, evaluate fun(query_points) through tf.map_fn, then
    restoring the leading dimensions.

    :param query_points: shape [...*, N, D]
    :param fun: callable that returns two tensors (e.g. a predict function)
    :return: two tensors (e.g. mean and variance) with shape [...*, ...]
    """

    leading_dim, query_points_flatten = _get_leading_dim_and_flatten(query_points)
    # leading_dim =...*, product = B
    # query_points_flatten: [B, N, D]

    mean_signature = tf.TensorSpec(None, query_points.dtype)
    var_signature = tf.TensorSpec(None, query_points.dtype)
    mean, var = tf.map_fn(
        fn=fun,
        elems=query_points_flatten,
        fn_output_signature=(mean_signature, var_signature),
    )  # [B, ..., L, N], [B, ..., L, N] (predict_f) or [B, ..., L, N, N] (predict_joint)

    return _restore_leading_dim(mean, leading_dim), _restore_leading_dim(var, leading_dim)


def _get_leading_dim_and_flatten(query_points: TensorType) -> tuple[TensorType, TensorType]:
    """
    :param query_points: shape [...*, N, D]
    :return: leading_dim = ....*, query_points_flatten, shape [B, N, D]
    """
    leading_dim = tf.shape(query_points)[:-2]  # =...*, product = B
    nd = tf.shape(query_points)[-2:]
    query_points_flatten = tf.reshape(query_points, (-1, nd[0], nd[1]))  # [B, N, D]
    return leading_dim, query_points_flatten


def _restore_leading_dim(x: TensorType, leading_dim: TensorType) -> TensorType:
    """
    "Un-flatten" the first dimension of x to leading_dim

    :param x: shape [B, ...]
    :param leading_dim: [...*]
    :return: shape [...*, ...]
    """
    single_x_shape = tf.shape(x[0])  # = [...]
    output_x_shape = tf.concat([leading_dim, single_x_shape], axis=0)  # = [...*, ...]
    return tf.reshape(x, output_x_shape)  # [...*, ...]



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains functionality for computing the non-dominated set
given a set of data points."""
from __future__ import annotations

import tensorflow as tf

from ...types import TensorType


def non_dominated(observations: TensorType) -> tuple[TensorType, TensorType]:
    """
    Computes the non-dominated set for a set of data points. Based on:
    https://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python

    If there are duplicate point(s) in the non-dominated set, this function will return
    as it is without removing the duplicate.

    :param observations: set of points with shape [N,D]
    :return: tf.Tensor of the non-dominated set [P,D] and a non-dominated point mask [N],
        P is the number of points in pareto front, the mask specifies whether each data point
        is non-dominated or not.
    """
    num_points = tf.shape(observations)[0]

    # Reordering the observations beforehand speeds up the search:
    mean = tf.reduce_mean(observations, axis=0)
    std = tf.math.reduce_std(observations, axis=0)
    weights = tf.reduce_sum(((observations - mean) / (std + 1e-7)), axis=1)
    sorting_indices = tf.argsort(weights)

    def cond(i: tf.Tensor, indices: tf.Tensor) -> tf.Tensor:
        return i < len(indices)

    def body(i: tf.Tensor, indices: tf.Tensor) -> tuple[tf.Tensor, tf.Tensor]:
        obs = tf.gather(observations, indices)
        nondominated = tf.reduce_any(obs < obs[i], axis=1) | tf.reduce_all(obs == obs[i], axis=1)
        i = tf.reduce_sum(tf.cast(nondominated[:i], tf.int32)) + 1
        indices = indices[nondominated]
        return i, indices

    _, indices = tf.while_loop(
        cond,
        body,
        loop_vars=(
            0,  # i
            tf.gather(tf.range(num_points), sorting_indices),  # indices
        ),
        shape_invariants=(
            tf.TensorShape([]),  # i
            tf.TensorShape([None]),  # indices
        ),
    )

    nondominated_observations = tf.gather(observations, indices)
    trues = tf.ones(tf.shape(indices), tf.bool)
    is_nondominated = tf.scatter_nd(indices[:, None], trues, [num_points])
    return nondominated_observations, is_nondominated



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod

from gpflow.base import Module
from gpflow.keras import tf_keras

from ...space import EncoderFunction
from ...types import TensorType
from ..interfaces import EncodedSupportsPredictY, SupportsGetObservationNoise
from ..optimizer import KerasOptimizer


class GPfluxPredictor(SupportsGetObservationNoise, EncodedSupportsPredictY, ABC):
    """
    A trainable wrapper for a GPflux deep Gaussian process model. The code assumes subclasses
    will use the Keras `fit` method for training, and so they should provide access to both a
    `model_keras` and `model_gpflux`.
    """

    def __init__(
        self, optimizer: KerasOptimizer | None = None, encoder: EncoderFunction | None = None
    ):
        """
        :param optimizer: The optimizer wrapper containing the optimizer with which to train the
            model and arguments for the wrapper and the optimizer. The optimizer must
            be an instance of a :class:`~tf.optimizers.Optimizer`. Defaults to
            :class:`~tf.optimizers.Adam` optimizer with 0.01 learning rate.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """
        if optimizer is None:
            optimizer = KerasOptimizer(tf_keras.optimizers.Adam(0.01))

        self._optimizer = optimizer
        self._encoder = encoder

    @property
    def encoder(self) -> EncoderFunction | None:
        return self._encoder

    @encoder.setter
    def encoder(self, encoder: EncoderFunction | None) -> None:
        self._encoder = encoder

    @property
    @abstractmethod
    def model_gpflux(self) -> Module:
        """The underlying GPflux model."""

    @property
    @abstractmethod
    def model_keras(self) -> tf_keras.Model:
        """Returns the compiled Keras model for training."""

    @property
    def optimizer(self) -> KerasOptimizer:
        """The optimizer wrapper for training the model."""
        return self._optimizer

    def predict_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """Note: unless otherwise noted, this returns the mean and variance of the last layer
        conditioned on one sample from the previous layers."""
        return self.model_gpflux.predict_f(query_points)

    @abstractmethod
    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        raise NotImplementedError

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        """Note: unless otherwise noted, this will return the prediction conditioned on one sample
        from the lower layers."""
        f_mean, f_var = self.model_gpflux.predict_f(query_points)
        return self.model_gpflux.likelihood_layer.likelihood.predict_mean_and_var(
            query_points, f_mean, f_var
        )

    def get_observation_noise(self) -> TensorType:
        """
        Return the variance of observation noise for homoscedastic likelihoods.

        :return: The observation noise.
        :raise NotImplementedError: If the model does not have a homoscedastic likelihood.
        """
        try:
            noise_variance = self.model_gpflux.likelihood_layer.likelihood.variance
        except AttributeError:
            raise NotImplementedError(f"Model {self!r} does not have scalar observation noise")

        return noise_variance



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Tuple, Union

import gpflow
import tensorflow as tf
import tensorflow_probability as tfp
from gpflow.keras import tf_keras

from ...data import Dataset
from ...types import TensorType
from ...utils import DEFAULTS
from ..optimizer import BatchOptimizer, Optimizer
from .interface import GPflowPredictor


def assert_data_is_compatible(new_data: Dataset, existing_data: Dataset) -> None:
    """
    Checks that new data is compatible with existing data.

    :param new_data: New data.
    :param existing_data: Existing data.
    :raise ValueError: if trailing dimensions of the query point or observation differ.
    """
    if new_data.query_points.shape[-1] != existing_data.query_points.shape[-1]:
        raise ValueError(
            f"Shape {new_data.query_points.shape} of new query points is incompatible with"
            f" shape {existing_data.query_points.shape} of existing query points. Trailing"
            f" dimensions must match."
        )

    if new_data.observations.shape[-1] != existing_data.observations.shape[-1]:
        raise ValueError(
            f"Shape {new_data.observations.shape} of new observations is incompatible with"
            f" shape {existing_data.observations.shape} of existing observations. Trailing"
            f" dimensions must match."
        )


def randomize_hyperparameters(object: gpflow.Module) -> None:
    """
    Sets hyperparameters to random samples from their prior distributions or (for Sigmoid
    constraints with no priors) their constrained domains. Note that it is up to the caller
    to ensure that the prior, if defined, is compatible with the transform.

    :param object: Any gpflow Module.
    """
    for param in object.trainable_parameters:
        if param.prior is not None:
            # handle constant priors for multi-dimensional parameters
            # Use python conditionals here to avoid creating tensorflow `tf.cond` ops,
            # i.e. using `len(param.shape)` instead of `tf.rank(param)`.
            # Otherwise, tensorflow generates repeating random sequences for hyperparameters, see
            # https://github.com/tensorflow/tensorflow/issues/61912.
            if param.prior.batch_shape == param.prior.event_shape == [] and len(param.shape) == 1:
                sample = param.prior.sample(tf.shape(param))
            else:
                sample = param.prior.sample()
            if param.prior_on is gpflow.base.PriorOn.UNCONSTRAINED:
                param.unconstrained_variable.assign(sample)
            else:
                param.assign(sample)

        elif isinstance(param.bijector, tfp.bijectors.Sigmoid):
            sample = tf.random.uniform(
                param.bijector.low.shape,
                minval=param.bijector.low,
                maxval=param.bijector.high,
                dtype=param.bijector.low.dtype,
            )
            param.assign(sample)


def squeeze_hyperparameters(
    object: gpflow.Module, alpha: float = 1e-2, epsilon: float = 1e-7
) -> None:
    """
    Squeezes the parameters to be strictly inside their range defined by the Sigmoid,
    or strictly greater than the limit defined by the Shift+Softplus.
    This avoids having Inf unconstrained values when the parameters are exactly at the boundary.

    :param object: Any gpflow Module.
    :param alpha: the proportion of the range with which to squeeze for the Sigmoid case
    :param epsilon: the value with which to offset the shift for the Softplus case.
    :raise ValueError: If ``alpha`` is not in (0,1) or epsilon <= 0
    """

    if not 0 < alpha < 1:
        raise ValueError(f"squeeze factor alpha must be in (0, 1), found {alpha}")

    if epsilon <= 0:
        raise ValueError(f"offset factor epsilon must be > 0, found {epsilon}")

    for param in object.trainable_parameters:
        if isinstance(param.bijector, tfp.bijectors.Sigmoid):
            delta = (param.bijector.high - param.bijector.low) * alpha
            squeezed_param = tf.math.minimum(param, param.bijector.high - delta)
            squeezed_param = tf.math.maximum(squeezed_param, param.bijector.low + delta)
            param.assign(squeezed_param)
        elif (
            isinstance(param.bijector, tfp.bijectors.Chain)
            and len(param.bijector.bijectors) == 2
            and isinstance(param.bijector.bijectors[0], tfp.bijectors.Shift)
            and isinstance(param.bijector.bijectors[1], tfp.bijectors.Softplus)
        ):
            if isinstance(param.bijector.bijectors[0], tfp.bijectors.Shift) and isinstance(
                param.bijector.bijectors[1], tfp.bijectors.Softplus
            ):
                low = param.bijector.bijectors[0].shift
                squeezed_param = tf.math.maximum(param, low + epsilon * tf.ones_like(param))
                param.assign(squeezed_param)


def check_optimizer(optimizer: Union[BatchOptimizer, Optimizer]) -> None:
    """
    Check that the optimizer for the GPflow models is using a correct optimizer wrapper.

    Stochastic gradient descent based methods implemented in TensorFlow would not
    work properly without mini-batches and hence :class:`~trieste.models.optimizers.BatchOptimizer`
    that prepares mini-batches and calls the optimizer iteratively needs to be used. GPflow's
    :class:`~gpflow.optimizers.Scipy` optimizer on the other hand should use the non-batch wrapper
    :class:`~trieste.models.optimizers.Optimizer`.

    :param optimizer: An instance of the optimizer wrapper with the underlying optimizer.
    :raise ValueError: If :class:`~tf.optimizers.Optimizer` is not using
        :class:`~trieste.models.optimizers.BatchOptimizer` or :class:`~gpflow.optimizers.Scipy` is
        using :class:`~trieste.models.optimizers.BatchOptimizer`.
    """
    if isinstance(optimizer.optimizer, gpflow.optimizers.Scipy):
        if isinstance(optimizer, BatchOptimizer):
            raise ValueError(
                f"""
                The gpflow.optimizers.Scipy can only be used with an Optimizer wrapper,
                however received {optimizer}.
                """
            )

    if isinstance(optimizer.optimizer, tf_keras.optimizers.Optimizer):
        if not isinstance(optimizer, BatchOptimizer):
            raise ValueError(
                f"""
                The tf.optimizers.Optimizer can only be used with a BatchOptimizer wrapper,
                however received {optimizer}.
                """
            )


def _covariance_between_points_for_variational_models(
    kernel: gpflow.kernels.Kernel,
    inducing_points: TensorType,
    q_sqrt: TensorType,
    query_points_1: TensorType,
    query_points_2: TensorType,
    whiten: bool,
) -> TensorType:
    r"""
    Compute the posterior covariance between sets of query points.

    .. math:: \Sigma_{12} = K_{1x}BK_{x2} + K_{12} - K_{1x}K_{xx}^{-1}K_{x2}

    where :math:`B = K_{xx}^{-1}(q_{sqrt}q_{sqrt}^T)K_{xx}^{-1}`
    or :math:`B = L^{-1}(q_{sqrt}q_{sqrt}^T)(L^{-1})^T` if we are using
    a whitened representation in our variational approximation. Here
    :math:`L` is the Cholesky decomposition of :math:`K_{xx}`.
    See :cite:`titsias2009variational` for a derivation.

    Note that this function can also be applied to
    our :class:`VariationalGaussianProcess` models by passing in the training
    data rather than the locations of the inducing points.

    Although query_points_2 must be a rank 2 tensor, query_points_1 can
    have leading dimensions.

    :inducing points: The input locations chosen for our variational approximation.
    :q_sqrt: The Cholesky decomposition of the covariance matrix of our
        variational distribution.
    :param query_points_1: Set of query points with shape [..., A, D]
    :param query_points_2: Sets of query points with shape [B, D]
    :param whiten:  If True then use whitened representations.
    :return: Covariance matrix between the sets of query points with shape [..., L, A, B]
        (L being the number of latent GPs = number of output dimensions)
    """

    tf.debugging.assert_shapes([(query_points_1, [..., "A", "D"]), (query_points_2, ["B", "D"])])

    num_latent = q_sqrt.shape[0]

    K, Kx1, Kx2, K12 = _compute_kernel_blocks(
        kernel, inducing_points, query_points_1, query_points_2, num_latent
    )

    L = tf.linalg.cholesky(K)  # [L, M, M]
    Linv_Kx1 = tf.linalg.triangular_solve(L, Kx1)  # [..., L, M, A]
    Linv_Kx2 = tf.linalg.triangular_solve(L, Kx2)  # [..., L, M, B]

    def _leading_mul(M_1: TensorType, M_2: TensorType, transpose_a: bool) -> TensorType:
        if transpose_a:  # The einsum below is just A^T*B over the last 2 dimensions.
            return tf.einsum("...lji,ljk->...lik", M_1, M_2)
        else:  # The einsum below is just A*B^T over the last 2 dimensions.
            return tf.einsum("...lij,lkj->...lik", M_1, M_2)

    if whiten:
        first_cov_term = _leading_mul(
            _leading_mul(Linv_Kx1, q_sqrt, transpose_a=True),  # [..., L, A, M]
            _leading_mul(Linv_Kx2, q_sqrt, transpose_a=True),  # [..., L, B, M]
            transpose_a=False,
        )  # [..., L, A, B]
    else:
        Linv_qsqrt = tf.linalg.triangular_solve(L, q_sqrt)  # [L, M, M]
        first_cov_term = _leading_mul(
            _leading_mul(Linv_Kx1, Linv_qsqrt, transpose_a=True),  # [..., L, A, M]
            _leading_mul(Linv_Kx2, Linv_qsqrt, transpose_a=True),  # [..., L, B, M]
            transpose_a=False,
        )  # [..., L, A, B]

    second_cov_term = K12  # [..., L, A, B]
    third_cov_term = _leading_mul(Linv_Kx1, Linv_Kx2, transpose_a=True)  # [..., L, A, B]
    cov = first_cov_term + second_cov_term - third_cov_term  # [..., L, A, B]

    tf.debugging.assert_shapes(
        [
            (query_points_1, [..., "N", "D"]),
            (query_points_2, ["M", "D"]),
            (cov, [..., "L", "N", "M"]),
        ]
    )
    return cov


def _compute_kernel_blocks(
    kernel: gpflow.kernels.Kernel,
    inducing_points: TensorType,
    query_points_1: TensorType,
    query_points_2: TensorType,
    num_latent: int,
) -> tuple[TensorType, TensorType, TensorType, TensorType]:
    """
    Return all the prior covariances required to calculate posterior covariances for each latent
    Gaussian process, as specified by the `num_latent` input.

    This function returns the covariance between: `inducing_points` and `query_points_1`;
    `inducing_points` and `query_points_2`; `query_points_1` and `query_points_2`;
    `inducing_points` and `inducing_points`.

    The calculations are performed differently depending on the type of
    kernel (single output, separate independent multi-output or shared independent
    multi-output) and inducing variables (simple set, SharedIndependent or SeparateIndependent).

    Note that `num_latents` is only used when we use a single kernel for a multi-output model.
    """

    if isinstance(kernel, (gpflow.kernels.SharedIndependent, gpflow.kernels.SeparateIndependent)):
        if isinstance(inducing_points, list):
            K = tf.concat(
                [ker(Z)[None, ...] for ker, Z in zip(kernel.kernels, inducing_points)], axis=0
            )
            Kx1 = tf.concat(
                [
                    ker(Z, query_points_1)[None, ...]
                    for ker, Z in zip(kernel.kernels, inducing_points)
                ],
                axis=0,
            )  # [..., L, M, A]
            Kx2 = tf.concat(
                [
                    ker(Z, query_points_2)[None, ...]
                    for ker, Z in zip(kernel.kernels, inducing_points)
                ],
                axis=0,
            )  # [L, M, B]
            K12 = tf.concat(
                [ker(query_points_1, query_points_2)[None, ...] for ker in kernel.kernels], axis=0
            )  # [L, M, B]
        else:
            K = kernel(inducing_points, full_cov=True, full_output_cov=False)  # [L, M, M]
            Kx1 = kernel(
                inducing_points, query_points_1, full_cov=True, full_output_cov=False
            )  # [..., L, M, A]
            Kx2 = kernel(
                inducing_points, query_points_2, full_cov=True, full_output_cov=False
            )  # [L, M, B]
            K12 = kernel(
                query_points_1, query_points_2, full_cov=True, full_output_cov=False
            )  # [..., L, A, B]
    else:  # simple calculations for the single output case
        K = kernel(inducing_points)  # [M, M]
        Kx1 = kernel(inducing_points, query_points_1)  # [..., M, A]
        Kx2 = kernel(inducing_points, query_points_2)  # [M, B]
        K12 = kernel(query_points_1, query_points_2)  # [..., A, B]

    if len(tf.shape(K)) == 2:  # if single kernel then repeat for all latent dimensions
        K = tf.repeat(tf.expand_dims(K, -3), num_latent, axis=-3)
        Kx1 = tf.repeat(tf.expand_dims(Kx1, -3), num_latent, axis=-3)
        Kx2 = tf.repeat(tf.expand_dims(Kx2, -3), num_latent, axis=-3)
        K12 = tf.repeat(tf.expand_dims(K12, -3), num_latent, axis=-3)
    elif len(tf.shape(K)) > 3:
        raise NotImplementedError(
            "Covariance between points is not supported for kernels of type {type(kernel)}."
        )

    tf.debugging.assert_shapes(
        [
            (K, ["L", "M", "M"]),
            (Kx1, ["L", "M", "A"]),
            (Kx2, ["L", "M", "B"]),
            (K12, ["L", "A", "B"]),
        ]
    )

    return K, Kx1, Kx2, K12


def _whiten_points(
    model: GPflowPredictor, inducing_points: TensorType
) -> Tuple[TensorType, TensorType]:
    """
    GPFlow's VGP and SVGP can use whitened representation, i.e.
    q_mu and q_sqrt parametrize q(v), and u = f(X) = L v, where L = cholesky(K(X, X))
    Hence we need to back-transform from f_mu and f_cov to obtain the updated
    new_q_mu and new_q_sqrt.

    :param model: The whitened model.
    :para inducing_points: The new inducing point locations.
    :return: The updated q_mu and q_sqrt with shapes [N, L] and [L, N, N], respectively.
    """

    f_mu, f_cov = model.model.predict_f(inducing_points, full_cov=True)  # [N, L], [L, N, N]
    f_mu -= model.model.mean_function(inducing_points)
    Knn = model.get_kernel()(inducing_points, full_cov=True)  # [N, N]
    jitter_mat = DEFAULTS.JITTER * tf.eye(tf.shape(inducing_points)[0], dtype=Knn.dtype)
    Lnn = tf.linalg.cholesky(Knn + jitter_mat)  # [N, N]
    new_q_mu = tf.linalg.triangular_solve(Lnn, f_mu)  # [N, L]
    tmp = tf.linalg.triangular_solve(Lnn[None], f_cov)  # [L, N, N], L⁻¹ f_cov
    S_v = tf.linalg.triangular_solve(Lnn[None], tf.linalg.matrix_transpose(tmp))  # [L, N, N]
    new_q_sqrt = tf.linalg.cholesky(S_v + jitter_mat)  # [L, N, N]

    return new_q_mu, new_q_sqrt



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping, Sequence
from typing import Callable, Optional

import tensorflow as tf

from ..data import Dataset
from ..models import ProbabilisticModelType
from ..types import Tag, TensorType
from .interface import AcquisitionFunction, AcquisitionFunctionBuilder


class Reducer(AcquisitionFunctionBuilder[ProbabilisticModelType]):
    r"""
    A :class:`Reducer` builds an :const:`~trieste.acquisition.AcquisitionFunction` whose output is
    calculated from the outputs of a number of other
    :const:`~trieste.acquisition.AcquisitionFunction`\ s. How these outputs are composed is defined
    by the method :meth:`_reduce`.
    """

    def __init__(self, *builders: AcquisitionFunctionBuilder[ProbabilisticModelType]):
        r"""
        :param \*builders: Acquisition function builders. At least one must be provided.
        :raise `~tf.errors.InvalidArgumentError`: If no builders are specified.
        """
        tf.debugging.assert_positive(
            len(builders), "At least one acquisition builder expected, got none."
        )

        self._acquisitions = builders

    def __repr__(self) -> str:
        """"""
        builders = ", ".join(map(repr, self._acquisitions))
        return f"{self.__class__.__name__}({builders})"

    def prepare_acquisition_function(
        self,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        r"""
        Return an acquisition function. This acquisition function is defined by first building
        acquisition functions from each of the
        :class:`~trieste.acquisition.AcquisitionFunctionBuilder`\ s specified at
        :meth:`__init__`, then reducing, with :meth:`_reduce`, the output of each of those
        acquisition functions.

        :param datasets: The data from the observer.
        :param models: The models over each dataset in ``datasets``.
        :return: The reduced acquisition function.
        """
        self.functions = tuple(
            acq.prepare_acquisition_function(models, datasets=datasets) for acq in self.acquisitions
        )

        def evaluate_acquisition_function_fn(at: TensorType) -> TensorType:
            return self._reduce_acquisition_functions(at, self.functions)

        return evaluate_acquisition_function_fn

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        models: Mapping[Tag, ProbabilisticModelType],
        datasets: Optional[Mapping[Tag, Dataset]] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param models: The model.
        :param datasets: Unused.
        """
        self.functions = tuple(
            acq.update_acquisition_function(function, models, datasets=datasets)
            for function, acq in zip(self.functions, self.acquisitions)
        )

        def evaluate_acquisition_function_fn(at: TensorType) -> TensorType:
            return self._reduce_acquisition_functions(at, self.functions)

        return evaluate_acquisition_function_fn

    @property
    def acquisitions(self) -> Sequence[AcquisitionFunctionBuilder[ProbabilisticModelType]]:
        """The acquisition function builders specified at class initialisation."""
        return self._acquisitions

    def _reduce_acquisition_functions(
        self, at: TensorType, acquisition_functions: Sequence[AcquisitionFunction]
    ) -> TensorType:
        return self._reduce([fn(at) for fn in acquisition_functions])

    @abstractmethod
    def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
        """
        :param inputs: The output of each constituent acquisition function.
        :return: The output of the reduced acquisition function.
        """
        raise NotImplementedError()


class Sum(Reducer[ProbabilisticModelType]):
    """
    :class:`Reducer` whose resulting acquisition function returns the element-wise sum of the
    outputs of constituent acquisition functions.
    """

    def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
        """
        :param inputs: The outputs of each acquisition function.
        :return: The element-wise sum of the ``inputs``.
        """
        return tf.add_n(inputs)


class Product(Reducer[ProbabilisticModelType]):
    """
    :class:`Reducer` whose resulting acquisition function returns the element-wise product of the
    outputs of constituent acquisition functions.
    """

    def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
        """
        :param inputs: The outputs of each acquisition function.
        :return: The element-wise product of the ``inputs``.
        """
        return tf.reduce_prod(inputs, axis=0)


class Map(Reducer[ProbabilisticModelType]):
    """
    :class:`Reducer` that accepts just one acquisition function builder and applies a
    given function to its output. For example ``Map(lambda x: -x, builder)`` would generate
    an acquisition function that returns the negative of the output of ``builder``.
    """

    def __init__(
        self,
        map_fn: Callable[[TensorType], TensorType],
        builder: AcquisitionFunctionBuilder[ProbabilisticModelType],
    ):
        """
        :param map_fn: Function to apply.
        :param builder: Acquisition function builder.
        """
        super().__init__(builder)
        self._map_fn = map_fn

    def _reduce(self, inputs: Sequence[TensorType]) -> TensorType:
        """
        :param inputs: The outputs of the acquisition function.
        :return: The result of applying the map function to ``inputs``.
        """
        tf.debugging.assert_equal(len(inputs), 1)
        return self._map_fn(inputs[0])
